import os

import src.utils as utils

def get_whitelist_object(pgm_object):
    key_yaml = pgm_object['structure']
    whitelist_object = read_in_whitelists(pgm_object['whitelist_path'])
    return whitelist_object


def read_in_whitelists(path):
    files = os.listdir(os.path.join(path, 'misc', 'json'))
    whitelists_object=test(files, {}, path)
    return whitelists_object


def test(files, whitelist_object, path):
    for file in files:
        if os.path.isfile(os.path.join(path, 'misc', 'json', file)) and not '.' in file:
            whitelist_object[file] = utils.read_whitelist(file, whitelist_path=path)
        elif os.path.isdir(os.path.join(path, 'misc', 'json', file)):
            new_files = [os.path.join(file, f) for f in os.listdir(os.path.join(path, 'misc', 'json', file))]
            whitelist_object = test(new_files, whitelist_object, path)
    return whitelist_object


def get_single_whitelist(ob, whitelist_object):
    """
    This functions returns a single whitelist of type 'plain' for a key that is
    specified within a given dictionary. If the organism is specified as well,
    dependent whitelists only contain the values for said organism.
    (-> used for metadata generation and editing) If no organism is given, the
    whitelists of multiple organisms are merged together. (-> used for
    searching)
    :param ob: a dictionary containing the key 'key_name' and optionally the
               key 'organism'
    :return: either a whitelist or None if no whitelist exists
    """

    # test if an organism was specified
    if 'organism' in ob:

        # save the organism in a dictionary to work like a result dict
        infos = {'organism': ob['organism']}

        # set all_plain to False since dependencies from the organism can be
        # taken into account
        all_plain = False

    else:

        # set an empty dictionary to work like a result dict -> no info was
        # given
        infos = {}

        # set all_plain to True since dependencies cannot be considered
        all_plain = True

    # read in the whitelist
    whitelist = utils.get_whitelist(ob['key_name'], infos, whitelist_object=whitelist_object, all_plain=all_plain)

    # test if the whitelist was found and read in correctly and return the list
    # of whitelist values
    if whitelist and 'whitelist' in whitelist:
        return whitelist['whitelist']

    # return None if no whitelist was found
    else:
        return None


def parse_whitelist(key_name, node, filled_object, whitelist_object):
    """
    This function read in a whitelist and parses it according to type, headers
    and whitelist keys
    :param key_name: the key for which a whitelist is defined
    :param node: the part of the read in metadata structure that contains the
                 key
    :param filled_object: a partially filled whitelist object
    :return: whitelist: the read in whitelist
             whitelist_type: the type of the whitelist
             input_type: the type of the input field in the web interface
             headers: a string containing headers
             whitelist_keys: the headings of a grouped whitelist
    """

    # initialize return values
    whitelist = None
    whitelist_type = None
    input_type = 'short_text'
    headers = None
    whitelist_keys = None
    double = []

    # whitelist is defined or special case merge
    if ('whitelist' in node and node['whitelist']) or (
            'special_case' in node and 'merge' in node['special_case']):

        # read in whitelist
        whitelist = utils.get_whitelist(key_name, filled_object, whitelist_object=whitelist_object)

        if whitelist is not None:

            if 'double' in whitelist:
                double = whitelist['double']

            # test if the right keys are present in the whitelist
            # -> format check
            if 'whitelist_type' in whitelist and 'whitelist' in whitelist:

                # set whitelist type , headers, whitelist_keys, whitelist and
                # input_type
                whitelist_type = whitelist['whitelist_type']
                headers = whitelist['headers'] if 'headers' in whitelist else None
                whitelist_keys = whitelist['whitelist_keys'] if \
                    'whitelist_keys' in whitelist else None
                whitelist = whitelist['whitelist']
                input_type = 'select'

                # TODO: remove?
                if whitelist_type == 'depend':
                    whitelist = None
                    input_type = 'dependable'

                # TODO: test if plain_group is already there
                elif whitelist_type == 'group':
                    if isinstance(whitelist, dict):
                        new_w = []
                        for key in whitelist:
                            new_w.append(
                                {'title': key,
                                 'whitelist': whitelist[key]})
                        input_type = 'group_select'
                        whitelist = new_w
                    else:
                        input_type = 'select'
                        whitelist_type = 'plain_group'

                # TODO: better solution for department
                # test if whitelist is longer than 30
                if whitelist and len(whitelist) > 30 and \
                        key_name not in ['department', 'factor']:

                    # set whitelist type to multi_autofill if it is a list
                    if node['list']:
                        input_type = 'multi_autofill'

                    # set whitelist type to single_autofill if it is a string
                    else:
                        input_type = 'single_autofill'

                    # set whitelist to None
                    # -> whitelists on the website will be called with
                    # get_single_whitelist function (from whitelist_parsing) and
                    # used with an autocompletion
                    # -> whitelist only gets send to website if the field is
                    # actually entered which saves space and time
                    whitelist = None

    # special case: value unit
    elif 'special_case' in node and 'value_unit' in node['special_case']:

        # read in unit whitelist and set input type to value_unit
        whitelist = utils.get_whitelist('unit', filled_object, whitelist_object=whitelist_object)
        if whitelist and 'whitelist' in whitelist:
            whitelist = whitelist['whitelist']
        input_type = 'value_unit'

    # boolean value
    elif node['input_type'] == 'bool':

        # set whitelist to True and False, input type to select and whitelist
        # type to plain
        whitelist = ['True', 'False']
        whitelist_type = 'plain'
        input_type = 'select'

    # no whitelist or special case
    else:

        # set input type as defines in general structure
        input_type = node['input_type']

    return whitelist, whitelist_type, input_type, headers, whitelist_keys, double



import argparse
import copy
import pathlib
import sys
import os
import time

sys.path.append(os.path.dirname(__file__))
from src.generate import Generate
from src import find_metafiles
from src import validate_yaml
from src import file_reading
from src import utils
from src import git_whitelists
from src.heatmap import create_heatmap



class FRED:

    def __init__(self, config):
        self.whitelist_repo, self.whitelist_branch, self.whitelist_path, \
        self.username, self.password, structure, self.update_whitelists, \
        self.output_path, self.filename, self.email = utils.parse_config(config)
        if self.whitelist_path == 'fred':
            self.whitelist_path = os.path.join(os.path.dirname(__file__), 'FRED_whitelists')
        else:
            self.whitelist_path = os.path.join(self.whitelist_path, 'FRED_whitelists')
        self.fetch_whitelists()
        self.structure = utils.read_in_yaml(structure)

    def fetch_whitelists(self):
        git_whitelists.get_whitelists(self.whitelist_path, self.whitelist_repo, self.whitelist_branch, self.update_whitelists)

    def find(self, search_path, search, output, output_filename, skip_validation):
        result = find_metafiles.find_projects(self.structure, search_path,
                                              search, True, skip_validation)
        if output == 'print':
            if len(result) > 0:

                # print summary of matching files
                print(find_metafiles.print_summary(result, output))

            else:

                # print information that there are no matching files
                print('No matches found')
        elif output == 'json':
            if not output_filename:
                output_filename = 'search_result'
            json_filename = f'{output_filename}.json'
            utils.save_as_json({'data': find_metafiles.print_summary(result, output)}, json_filename)
            print(
                f'The report was saved to the file \'{json_filename}\'.')

    def generate(self, path, project_id, mandatory_only):
        gen = Generate(path, project_id, mandatory_only, self.filename, self.structure, self.email)
        gen.generate()

    def validate(self, logical_validation, path, output, output_filename, save_empty=False):
        validation_reports = {'all_files': 1,
                              'corrupt_files': {'count': 0, 'report': []},
                              'error_count': 0, 'warning_count': 0}
        if os.path.isdir(path):
            metafiles, validation_reports = file_reading.iterate_dir_metafiles(
                self.structure, [path], filename=self.filename,
                logical_validation=logical_validation, yaml=copy.deepcopy(self.structure),
                whitelist_path=self.whitelist_path)
        else:
            metafile = utils.read_in_yaml(path)
            file_reports = {'file': metafile, 'error': None, 'warning': None}
            valid, missing_mandatory_keys, invalid_keys, \
            invalid_entries, invalid_values, logical_warn = validate_yaml.validate_file(
                metafile, self.structure, self.filename,
                logical_validation=logical_validation, yaml=copy.deepcopy(self.structure),
                whitelist_path=self.whitelist_path)
            metafile['path'] = path
            if not valid:
                validation_reports['corrupt_files']['count'] = 1
                validation_reports['error_count'] += (
                            len(missing_mandatory_keys) + len(
                        invalid_keys) + len(invalid_entries) + len(
                        invalid_values))
                file_reports['error'] = (
                missing_mandatory_keys, invalid_keys, invalid_entries,
                invalid_values)
            if len(logical_warn) > 0:
                validation_reports['corrupt_files']['count'] = 1
                validation_reports['warning_count'] += len(logical_warn)
                file_reports['warning'] = logical_warn
            validation_reports['corrupt_files']['report'].append(file_reports)

        print(f'{validation_reports["all_files"]} files were validated.')
        print(
            f'Found {validation_reports["error_count"]} errors and {validation_reports["warning_count"]} warnings in {validation_reports["corrupt_files"]["count"]} of those files.')

        if validation_reports['corrupt_files']['count'] > 0 or save_empty is True:

            res = None
            if output is not None:
                if output == 'print':
                    res = ['print report']
                elif output == 'txt':
                    res = ['save report to txt file']
                elif output == 'json':
                    res = ['save report to json file']
                elif output == 'yaml':
                    res = ['save report to yaml file']
            else:
                options = ['print report', 'save report to txt file',
                           'save report to json file',
                           'save report to yaml file']
                print(
                    f'Do you want to see a report? Choose from the following options (1,...,{len(options)} or n)')
                ask = Generate('', '', False, self.filename, self.structure, self.email)
                ask.print_option_list(options, '')
                res = ask.parse_input_list(options, True)

            try:
                output_report = {
                    'report': copy.deepcopy(validation_reports)['corrupt_files']['report']}
            except KeyError:
                output_report = {'report': []}
            for elem in output_report['report']:
                id = list(utils.find_keys(elem['file'], 'id'))
                if len(id) > 0:
                    elem['id'] = id[0]
                else:
                    elem['id'] = 'missing'
                elem['path'] = elem['file']['path']
                errors = list(elem['error']) if 'error' in elem and elem[
                    'error'] is not None else []
                elem['error'] = {}
                elem.pop('file')
                for i in range(len(errors)):

                    if len(errors[i]) > 0:
                        if i == 0:
                            elem['error']['missing_mandatory_keys'] = errors[i]
                        elif i == 1:
                            elem['error']['invalid_keys'] = errors[i]
                        elif i == 2:
                            whitelist_values = []
                            for v in errors[i]:
                                key = ':'.join(v.split(':')[:-1])
                                entry = v.split(':')[-1]
                                whitelist_values.append(
                                    entry + ' in ' + key + '\n')
                            elem['error']['invalid_entries'] = whitelist_values
                        elif i == 3:
                            value = []
                            for v in errors[i]:
                                value.append(f'{v[0]}: {v[1]} -> {v[2]}')
                            elem['error']['invalid_values'] = value

                if 'warning' in elem:
                    if elem['warning'] is not None:
                        for i in range(len(elem['warning'])):
                            elem['warning'][
                                i] = f'{elem["warning"][i][0]}: {elem["warning"][i][1]}'
                    else:
                        elem.pop('warning')

            if res is not None:
                if output_filename is None:
                    timestamp = time.time()
                    output_filename = f'validation_report_{str(timestamp).split(".")[0]}'

                rep = ''
                for report in validation_reports['corrupt_files']['report']:
                    rep += f'{"".center(80, "_")}\n\n'
                    rep += validate_yaml.print_full_report(report['file'],
                                                           report['error'],
                                                           report['warning'])
                rep += f'{"".center(80, "_")}\n\n'

                if 'save report to txt file' in res:
                    txt_filename = f'{output_filename}.txt'
                    txt_f = open(txt_filename, 'w')
                    txt_f.write(rep)
                    print(
                        f'The report was saved to the file \'{txt_filename}\'.')
                    txt_f.close()

                if 'save report to json file' in res:
                    json_filename = f'{output_filename}.json'
                    utils.save_as_json(output_report, json_filename)
                    print(
                        f'The report was saved to the file \'{json_filename}\'.')

                if 'save report to yaml file' in res:
                    yaml_filename = f'{output_filename}.yaml'
                    utils.save_as_yaml(output_report, yaml_filename)
                    print(
                        f'The report was saved to the file \'{yaml_filename}\'.')

                if 'print report' in res:
                    print(rep)

        return validation_reports["error_count"], validation_reports["warning_count"]

    #def edit(self, path, mandatory_only):
    #    try:
    #        size = os.get_terminal_size()
    #        size = size.columns
    #    except OSError:
    #        size = 80

    #    edit_file.edit_file(path, self.filename, mandatory_only, size)

    def add_value(self, path, position, value, edit_existing):
        files, errors = file_reading.iterate_dir_metafiles(self.structure,
                                                           [path],
                                                           self.filename, False,
                                                           return_false=True,
                                                           whitelist_path=
                                                           self.whitelist_path)
        position = position.split(':')
        # TODO: type
        for file in files:
            file = utils.add_value_at_pos(self.structure, file, position, value,
                                          edit_existing)
            save_path = file['path']
            file.pop('path')
            print(f'edited file {save_path}')
            utils.save_as_yaml(file, save_path)

def find(args):
    """
    calls script find_metafiles to find matching files and print results
    :param args:
        path: a path of a folder that should be searched for metadata files
        search: a string specifying search parameters linked via 'and', 'or'
                and 'not'
    """
    finding = FRED(args.config)
    finding.find(args.path, args.search, args.output, args.filename,
                 args.skip_validation)


def generate(args):
    """
    calls script generate_metafile to start dialog
    :param args:
    """
    generating = FRED(args.config)
    generating.generate(args.path, args.id, args.mandatory_only)


def validate(args):
    validating = FRED(args.config)
    errors, warnings = validating.validate(not args.skip_logic, args.path, args.output, args.filename)


def plot(args):
    fred_object = FRED(args.config)
    input_file = utils.read_in_yaml(args.path)
    plots = create_heatmap.get_heatmap(input_file, fred_object.structure, mode=args.mode, labels=args.labels, background=args.background, sample_labels=args.sample_labels, condition_labels=args.condition_labels, transpose=args.transpose, drop_defaults=args.drop_defaults)
    output_filename = args.filename if args.filename is not None else 'fig1'
    if len(plots) > 0:
        try:
            plot = plots[args.setting-1][1]
        except IndexError:
            print(f'Setting exp{args.setting} does not exist. Defaulting to exp1.')
            plot = plots[0][1]
        
        if plot is not None:
            if args.output == 'png':
                plot.write_image(f'{output_filename}.{args.output}', format='png')
                print(f'Plot was saved to {output_filename}.{args.output}')
            elif args.output == 'html':
                with open(f'{output_filename}.{args.output}', 'w') as file:
                    file.write(plot.to_html(full_html=False, include_plotlyjs='cdn'))
                print(f'Plot was saved to {output_filename}.{args.output}')
            else:
                plot.show()
        else:
            print('Plot could not be created due to missing samples or conditions.')
    else:
        print('No settings found.')

#def edit(args):
#    editing = FRED(args.config)
#    editing.edit(args.path, args.mandatory_only)

def add_value(args):
    adding = FRED(args.config)
    adding.add_value(args.path, args.position, args.value, args.edit_existing)

def main():

    parser = argparse.ArgumentParser(prog='metaTools.py')
    subparsers = parser.add_subparsers(title='commands')

    find_function = subparsers.add_parser('find',
                                          help='This command is used to find '
                                               'projects by searching the '
                                               'metadata files.')

    find_group = find_function.add_argument_group('mandatory arguments')
    find_group.add_argument('-p', '--path', type=pathlib.Path, required=True,
                               help='The path to be searched')
    find_group.add_argument('-s', '--search', type=str, required=True,
                               help='The search parameters')
    find_group.add_argument('-c', '--config', type=pathlib.Path,
                              help='Config file', default=os.path.join(os.path.dirname(__file__), 'config', 'config.yaml'))
    find_group.add_argument('-o', '--output', default='print', choices=['json', 'print'])
    find_group.add_argument('-f', '--filename', default=None)
    find_group.add_argument('-sv', '--skip_validation', default=False,
                            action='store_true')
    find_function.set_defaults(func=find)

    create_function = subparsers.add_parser('generate',
                                            help='This command is used to '
                                                 'create a metadata file.')
    create_group = create_function.add_argument_group('mandatory arguments')
    create_group.add_argument('-p', '--path', type=pathlib.Path,
                                 required=True,
                                 help='The path to save the yaml')
    create_group.add_argument('-c', '--config', type=pathlib.Path,
                              help='Config file', default=os.path.join(os.path.dirname(__file__), 'config', 'config.yaml'))
    create_group.add_argument('-id', '--id', type=str,
                                 required=True,
                                 help='The ID of the experiment')
    create_function.add_argument('-mo', '--mandatory_only', default=False,
                                 action='store_true',
                                 help='If True, only mandatory keys will '
                                      'be filled out')
    create_function.add_argument('-m', '--mode', default='metadata', choices=['metadata', 'mamplan'])
    create_function.set_defaults(func=generate)

    validate_function = subparsers.add_parser('validate',
                                              help='')
    validate_group = validate_function.add_argument_group('mandatory arguments')
    validate_group.add_argument('-p', '--path', type=pathlib.Path, required=True)
    validate_function.add_argument('-l', '--skip_logic', default=False,
                                   action='store_true')
    validate_function.add_argument('-m', '--mode', default='metadata', choices=['metadata', 'mamplan'])
    validate_function.add_argument('-o', '--output', default=None, choices=['json', 'txt', 'print', 'yaml'])
    validate_function.add_argument('-f', '--filename', default=None)
    validate_function.add_argument('-c', '--config', type=pathlib.Path,
                              help='Config file', default=os.path.join(os.path.dirname(__file__), 'config', 'config.yaml'))
    validate_function.set_defaults(func=validate)

    #edit_function = subparsers.add_parser('edit', help='')
    #edit_group = edit_function.add_argument_group('mandatory_arguments')
    #edit_group.add_argument('-p', '--path', type=pathlib.Path, required=True)
    #edit_function.add_argument('-mo', '--mandatory_only', default=False,
    #                             action='store_true',
    #                             help='If True, only mandatory keys will '
    #                                  'be filled out')
    #edit_function.add_argument('-c', '--config', type=pathlib.Path,
    #                          help='Config file', default='config.yaml')
    #edit_function.add_argument('-m', '--mode', default='metadata', choices=['metadata', 'mamplan'])
    #edit_function.set_defaults(func=edit)

    add_value_function = subparsers.add_parser('add_value', help='')
    add_value_function.add_argument('-m', '--mode', default='metadata', choices=['metadata', 'mamplan'])
    add_value_function.add_argument('-pos', '--position', required=True)
    add_value_function.add_argument('-v', '--value', required=True)
    add_value_function.add_argument('-t', '--type', default='str', choices=['str', 'int', 'float', 'bool'])
    add_value_function.add_argument('-p', '--path', type=pathlib.Path, required=True)
    add_value_function.add_argument('-e', '--edit_existing', default=False, action='store_true')
    add_value_function.add_argument('-c', '--config', type=pathlib.Path,
                              help='Config file', default=os.path.join(os.path.dirname(__file__), 'config', 'config.yaml'))
    add_value_function.set_defaults(func=add_value)

    plot_function = subparsers.add_parser('plot', help='')
    plot_function.add_argument('-p', '--path', type=pathlib.Path, required=True)
    plot_function.add_argument('-c', '--config', type=pathlib.Path,
                              help='Config file', default=os.path.join(os.path.dirname(__file__), 'config', 'config.yaml'))
    plot_function.add_argument('-m', '--mode', default='samples', choices=['samples', 'conditions'])
    plot_function.add_argument('-s', '--setting', type=int, default=1)
    plot_function.add_argument('-l', '--labels', default='factors', choices=['factors', 'all', 'none'])
    plot_function.add_argument('-o', '--output', default='show', choices=['show', 'png', 'html', 'dash'])
    plot_function.add_argument('-f', '--filename', type=pathlib.Path)
    plot_function.add_argument('-b', '--background', default=False, action='store_true', 
                               help='If stated, the background will be displayed in white. Per default it is transparent.')
    plot_function.add_argument('-cl', '--condition_labels', default=False, action='store_true', 
                               help='If stated, the label of the condition will be displayed as a name. Per default an index is stated. ')
    plot_function.add_argument('-sl', '--sample_labels', default=False, action='store_true', 
                               help='If stated, the label of the sample will be displayed as a name. Per default an index is stated. ')
    plot_function.add_argument('-t', '--transpose', default=False, action='store_true', 
                               )
    plot_function.add_argument('-d', '--drop_defaults', default=False, action='store_true', 
                               )
    plot_function.set_defaults(func=plot)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError:
        parser.print_help()


if __name__ == "__main__":

    main()



#!/usr/bin/env python

import os
import sys
import yaml.parser
import yaml.scanner

sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'fred', 'src'))
import utils

#TODO: test special_case

def test_key_yaml(all_keys, inner_keys):
    try:
        key_yaml = utils.read_in_yaml(os.path.join(os.path.dirname(__file__), '..', 'fred', 'structure', 'keys.yaml'))
    except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
        print(f'The yaml file could not be parsed:\n{e}')
        sys.exit(1)

    for key in key_yaml:
        test_properties(key_yaml[key], key, all_keys, inner_keys)


def test_properties(part, key, all_keys, inner_keys):
    for prop in all_keys:
        if prop not in part:
            print(f'Property \'{prop}\' is missing for key \'{key}\'')
            sys.exit(1)
        else:
            error, prop_type = test_property_input(part[prop], prop, all_keys)
            if error:
                print(f'Wrong input \'{part[prop]}\' for property \'{prop}\' for key \'{key}\'. Input should be {prop_type}')
                sys.exit(1)

    if isinstance(part['value'], dict):
        for sub_key in part['value']:
            test_properties(part['value'][sub_key], f'{key}:{sub_key}', all_keys, inner_keys)
    else:
        test_for_inner_keys(part, key, inner_keys)


def test_for_inner_keys(part, key, inner_keys):
    for prop in inner_keys:
        if prop not in part:
            print(f'Property \'{prop}\' is missing for key \'{key}\'')
            sys.exit(1)
        else:
            error, prop_type = test_property_input(part[prop], prop, inner_keys)
            if error:
                print(
                    f'Wrong input \'{part[prop]}\' for property \'{prop}\' for key \'{key}\'. Input should be {prop_type}')
                sys.exit(1)


def test_property_input(prop_input, prop, keys):
    if keys[prop] == 'bool':
        if type(prop_input) != bool:
            return True, 'of type \'bool\''
    elif keys[prop] == 'str':
        if type(prop_input) != str:
            if prop != 'desc':
                return True, 'of type \'str\''
    elif keys[prop] == 'int':
        if type(prop_input) != int:
            return True, 'of type \'int\''
    elif keys[prop] == 'value':
        if prop_input is not None and not isinstance(prop_input, dict) and type(prop_input) not in [str, bool, int]:
            return True, 'either a dictionary or a default value that can be \'None\' or of type \'str\', \'bool\' or \'int\''
    elif isinstance(keys[prop], list):
        if prop_input not in keys[prop]:
            return True, ', '.join(f'\'{x}\'' for x in keys[prop])
    return False, 'None'


if __name__ == "__main__":

    all_keys = {'mandatory': 'bool',
                'list': 'bool',
                'display_name': 'str',
                'desc': 'str',
                'value': 'value'}

    inner_keys = {'whitelist': 'bool',
                  'input_type': ['short_text', 'long_text', 'select', 'number', 'bool', 'date', 'restricted_short_text']}

    test_key_yaml(all_keys, inner_keys)

    sys.exit(0)



import src.validate_yaml as validate_yaml
import src.web_interface.wi_utils as wi_utils
import src.web_interface.wi_object_to_yaml as oty
import src.utils as utils

# TODO: rewrite and documentation


def validate_object(wi_object, key_yaml, read_in_whitelists, finish, email):
    """
    This function performs a validation over the wi object
    :param wi_object: the filled wi object
    :return: validation_object: the validated wi object with errors and
                                warnings
    """
    pooled = None
    organisms = []
    settings = []
    for setting in wi_object['experimental_setting']['list_value']:
        for elem in setting:
            if elem['position'].split(':')[-1] == 'organism':
                if isinstance(elem['value'], dict):
                    organisms.append(elem['value']['organism_name'].split(
                        ' ')[0])
                else:
                    organisms.append(elem['value'].split(' ')[0])
            elif elem['position'].split(':')[-1] == 'setting_id':
                settings.append(elem['value'])
    warnings = {}
    errors = {}
    pubmed_ids = []

    for part in ['project', 'experimental_setting', 'technical_details']:

        part, wi_object[part], pooled, organisms, warnings[part], \
                errors[part], pubmed_ids = validate_part(part, wi_object[part], [], pooled,
                                             organisms, settings, [], pubmed_ids)

    pubmed_ids = list(set(pubmed_ids))
    if any(x not in wi_object['publication_records'] for x in pubmed_ids):
        wi_object['publication_records'] = get_publications(
            wi_object['publication_records'], pubmed_ids, email)

    if finish:
        parsed = oty.parse_object(wi_object, key_yaml, read_in_whitelists)
        for part in ['project', 'experimental_setting', 'technical_details']:
            if part not in parsed:
                parsed[part] = {}
            valid, missing_mandatory_keys, invalid_keys, invalid_entries, \
            invalid_value, logical_warn = validate_yaml.validate_file(parsed[part], key_yaml[part]['value'], 'metadata', logical_validation=False, generated=False, only_mandatory=True)
            for elem in missing_mandatory_keys:
                diaplay_name = list(utils.find_keys(key_yaml, elem.split(':')[-1]))[0]['display_name']
                errors[part].append({'position': f'{part}:{elem}',
                                     'title': f'Missing {diaplay_name}',
                                     'message': 'Missing mandatory input'})
    validation_object = {'object': wi_object, 'errors': errors,
                         'warnings': warnings}
    return validation_object


def get_publications(publications, pubmed_ids, email):
    pubmed_entries = utils.get_publication_object(pubmed_ids, email)
    for pubmed_entry in pubmed_entries:
        if pubmed_entry['Id'] not in publications:
            publications[pubmed_entry['Id']] = {
                'pubmed_id': int(pubmed_entry['Id'])}

            try:
                publications[pubmed_entry['Id']]['title'] = str(pubmed_entry['Title'])
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['year'] = int(pubmed_entry['PubDate'].split(' ')[0])
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['author'] = \
                    [str(x) for x in pubmed_entry['AuthorList'][0:3]] + \
                    ['et. al'] if len(pubmed_entry['AuthorList']) > 3 else \
                        [str(x) for x in pubmed_entry['AuthorList']]
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['journal'] = \
                    str(pubmed_entry['Source'])
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['volume'] = \
                    int(pubmed_entry['Volume'])
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['issue'] = \
                    int(pubmed_entry['Issue'])
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['pages'] = \
                    str(pubmed_entry['Pages'])
            except ValueError:
                pass

            try:
                publications[pubmed_entry['Id']]['doi'] = \
                    str(pubmed_entry['DOI'])
            except ValueError:
                pass

    return publications


def validate_part(elem, wi_object, warnings, pooled, organisms, settings, errors, pubmed_ids):
    """
    This function is used to validate a part of the WI object. If an error or a
    warning is found than it is added to the description of the key whose value
    contains the error/warning. This function is called simultaneously for the
    parts 'project', 'experimental_setting' and 'technical_details'.
    :param elem: the key of the part that is being validated
    :param wi_object: the object that is validated
    :param warnings: a list containing all warnings
    :param pooled: a boolean that states if the samples were pooled
    :param organisms: a list of all contained organisms
    :param errors: a list containing all errors
    :return: elem: the key of the part that is being validated
             wi_object: the validated object containing error and warn messages
             pooled: a boolean stating if the samples are pooled
             organisms: a list of all contained organisms
             warnings: a list containing all found warnings
             errors: a list containing all found errors
    """
    error_desc = ''
    warning_desc = ''
    if isinstance(wi_object, dict):
        if 'desc' in wi_object and 'backup_desc' not in wi_object:
            wi_object['backup_desc'] = wi_object['desc']

        if wi_object['list']:
            if wi_object['position'].split(':')[-1] == 'techniques':
                tech_settings = []
                for t_elem in wi_object['list_value']:
                    for t_sub_elem in t_elem:
                        if t_sub_elem['position'].split(':')[-1] == 'setting':
                            tech_settings.append(t_sub_elem['value'])
                warning, warn_text = validate_yaml.validate_techniques(settings, tech_settings)
                if warning:
                    warnings.append({'position': wi_object['position'],
                                     'title': 'Invalid Technique',
                                     'message': warn_text})

                    warn_text = warn_text.replace('\n', '<br>')
                    warning_desc = \
                        f'{warning_desc}' \
                        f'{"<br>" if warning_desc != "" else ""}' \
                        f'<font color="orange">{warn_text}</font>'
                    wi_object['desc'] =  \
                        f'{wi_object["backup_desc"]}' \
                        f'{"<br>" if wi_object["backup_desc"] != "" else ""}' \
                        f'{warning_desc}'
                else:
                    wi_object['desc'] = wi_object['backup_desc']
            elif not any([isinstance(x, dict) or isinstance(x, list) for x in
                        wi_object['list_value']]):
                error = False
                messages = []
                for sub_elem in wi_object['list_value']:
                    valid, message = validate_yaml.validate_value(
                        sub_elem, wi_object['input_type'],
                        wi_object['position'].split(':')[-1])
                    if not valid:
                        error = True
                        messages.append((sub_elem, message))
                        errors.append({'position': wi_object['position'],
                                       'title': f'Invalid {wi_object["displayName"]}',
                                       'message': message})
                wi_object['error'] = error
                if error:
                    message = ', '.join(
                        [f'{msg[0]}: {msg[1]}' for msg in messages])
                    error_desc = f'{error_desc}' \
                                 f'{"<br>" if error_desc != "" else ""}' \
                                 f'<font color="red">{message}</font>'
                wi_object['desc'] = \
                    f'{wi_object["backup_desc"]}' \
                    f'{"<br>" if wi_object["backup_desc"] != "" else ""}' \
                    f'{error_desc}{"<br>" if error_desc != "" else ""}' \
                    f'{warning_desc}'
            else:
                elem, wi_object[
                    'list_value'], pooled, organisms, warnings, errors, pubmed_ids = \
                    validate_part(
                        elem, wi_object['list_value'], warnings, pooled,
                        organisms, settings, errors, pubmed_ids)
        else:
            if 'input_fields' in wi_object:
                elem, wi_object[
                    'input_fields'], pooled, organisms, warnings, errors, pubmed_ids = \
                    validate_part(
                        elem, wi_object['input_fields'], warnings, pooled,
                        organisms, settings, errors, pubmed_ids)
            else:
                if wi_object['position'].split(':')[-1] == 'pubmed_id' and wi_object['value'] is not None:
                    pubmed_ids.append(wi_object['value'])
                if wi_object['value'] is not None and wi_object['value'] != '':
                    if wi_object['input_type'] == 'date':
                        value = wi_utils.date_to_str(wi_object['value'])
                    else:
                        value = wi_object['value']
                    valid, message = validate_yaml.validate_value(
                        value, wi_object['input_type'],
                        wi_object['position'].split(':')[-1])
                    wi_object['error'] = not valid
                    if not valid:
                        errors.append({'position': wi_object["position"],
                                       'title': f'Invalid {wi_object["displayName"]}',
                                       'message': message})
                        error_desc = f'{error_desc}' \
                                     f'{"<br>" if error_desc != "" else ""}' \
                                     f'<font color="red">{message}</font>'
                    warning = False
                    warn_text = None
                    warn_title = ''
                    key = wi_object['position'].split(':')[-1]
                    if key == 'pooled':
                        pooled = wi_object['value']
                    elif key == 'donor_count':
                        warning, warn_text = \
                            validate_yaml.validate_donor_count(
                                pooled, wi_object['value'])
                        warn_title = 'Illogical donor count'
                    elif key == 'organism':
                        organisms.append(wi_object['value'].split(' ')[0])
                    elif key == 'reference_genome':
                        warning, warn_text = \
                            validate_yaml.validate_reference_genome(
                                organisms, wi_object['value'])
                        warn_title = 'Illogical reference genome'
                    wi_object['warning'] = warning
                    if warning:
                        warnings.append({'position': wi_object["position"],
                                         'title': warn_title,
                                         'message': warn_text})
                        warning_desc = \
                            f'{warning_desc}' \
                            f'{"<br>" if warning_desc != "" else ""}' \
                            f'<font color="orange">{warn_text}</font>'
                    wi_object['desc'] = \
                        f'{wi_object["backup_desc"]}' \
                        f'{"<br>" if wi_object["backup_desc"] != "" else ""}' \
                        f'{error_desc}{"<br>" if error_desc != "" else ""}' \
                        f'{warning_desc}'
    elif isinstance(wi_object, list):
        for i in range(len(wi_object)):
            elem, wi_object[
                i], pooled, organisms, warnings, errors, pubmed_ids = validate_part(
                elem, wi_object[i], warnings, pooled, organisms, settings,
                errors, pubmed_ids)
    return elem, wi_object, pooled, organisms, warnings, errors, pubmed_ids



import src.utils as utils
import src.web_interface.whitelist_parsing as whitelist_parsing
import src.web_interface.yaml_to_wi_object as yto
import src.web_interface.wi_utils as wi_utils
import copy
import re


def get_factors(organism, key_yaml, read_in_whitelists):
    """
    This function returns all experimental factors with a whitelist of their
    values
    :param key_yaml: the read in general structure
    :param organism: the organism that was selected by the user
    :return: factor_value: a dictionary containing factors and whitelists
    """

    factor_desc = list(utils.find_keys(key_yaml, 'experimental_factors'))[0][
        'factor_desc']
    f_node = list(utils.find_keys(key_yaml, 'factor'))[0]
    # initialize dictionary with all factors
    factor_list, whitelist_type, input_type, headers, whitelist_keys, double = \
        whitelist_parsing.parse_whitelist(
            'factor', f_node, {'organism': organism}, whitelist_object=read_in_whitelists)

    factor_value = {'factor': factor_list, 'desc': factor_desc}
    plain_factors = []

    for elem in factor_list:
        plain_factors += elem['whitelist']
    # initialize empty dictionary to store values of each factor
    values = {}

    # iterate over factors
    for factor in plain_factors:

        # get attributes of factor from general structure
        node = list(utils.find_keys(key_yaml, factor))

        # factor was found in general structure
        if len(node) > 0:

            # call function 'get_factor_values' to get whitelist information
            whitelist, whitelist_type, input_type, headers, w_keys, double, nested_infos = \
                get_factor_values(factor, node[0], {'organism': organism}, read_in_whitelists)

            # change single_autofill to multi_autofill because all factors can
            # have multiple values
            if input_type == 'single_autofill':
                input_type = 'multi_autofill'

            # save whitelist, input_type and whitelist_type for the values of
            # one factor
            values[factor] = {'whitelist': whitelist,
                              'input_type': input_type,
                              'whitelist_type': whitelist_type,
                              'displayName': node[0]['display_name'],
                              'desc': utils.print_desc(
                                  node[0]['factor_desc'] if 'factor_desc' in
                                                            node[0] else
                                  node[0]['desc'], 'html').replace('\n',
                                                                   '<br>')}

            # add search_info if input is of type single- or multi-autofill
            if input_type == 'multi_autofill':
                values[factor]['search_info'] = {'organism': organism,
                                                 'key_name': factor}

            if input_type == 'restricted_short_text':
                # default
                values[factor]['restriction'] = {'regex': ' ',
                                                 'max_length': 10}

                if 'special_case' in node and 'restriction' in node[
                    'special_case']:
                    if 'regex' in node['special_case']['restriction']:
                        values[factor]['restriction']['regex'] = \
                            node['special_case']['restriction']['regex']
                    if 'max_length' in node['special_case']['restriction']:
                        values[factor]['restriction']['max_length'] = \
                            node['special_case']['restriction']['max_length']

            # add header and whitelist keys if they are defined
            if headers is not None:
                values[factor]['headers'] = headers
            if w_keys is not None:
                values[factor]['whitelist_keys'] = w_keys
            if len(double) > 0:
                values[factor]['double'] = double
            if nested_infos is not None:
                values[factor]['nested_infos'] = nested_infos

    # add the values to the dictionary
    factor_value['values'] = values

    return factor_value


def get_factor_values(key, node, filled_object, read_in_whitelists, nested_infos=None):
    """
    This function is used to get the whitelists of experimental factors
    including the whitelist type, input type, headers and whitelist keys
    :param key: the name of the experimental factor
    :param node: the part of the general structure containing information of
                 the factor
    :param filled_object: a dictionary storing the used organism and used for
                          parsing of whitelists of type 'depend' (dependent
                          on organism)
    :return:
    whitelist: a list or dictionary of available values for input (type depends
               on whitelist_type)
    whitelist_type: the type of the whitelist (e.g. plain, group, ...)
    input_type: the input type for the input field in the web interface (e.g.
                short_text, select, single_autofill, ...)
    headers: the headers of the whitelist (None if no headers are defined)
    w_keys: the keys of a whitelist of type group that was rewritten to type
            plain (None if the whitelist is not plain group)
    """

    # initialize headers, whitelist keys and whitelist type with None
    headers = None
    w_keys = None
    whitelist_type = None
    double = []

    # value is a dictionary and no special case
    if isinstance(node['value'], dict) and not (
            'special_case' in node and ('merge' in node['special_case'] or
                                        'value_unit' in node['special_case'])):

        # initialize whitelist as empty list
        whitelist = []
        nested_infos = {}

        # iterate over the keys of the value
        for k in node['value']:

            key_info = {}

            # initialize an empty dictionary to store the properties of
            # the key
            k_val = {}

            # call this function to get the whitelist information for the keys
            k_val['whitelist'], k_val['whitelist_type'], k_val['input_type'], \
            header, whitelist_keys, doubled, nested_infos = get_factor_values(
                k, node['value'][k],
                filled_object, read_in_whitelists, nested_infos=nested_infos)

            # add header and whitelist keys to dictionary if they are defined
            if header is not None:
                k_val['headers'] = header
                key_info['headers'] = header
            if whitelist_keys is not None:
                k_val['whitelist_keys'] = whitelist_keys
                key_info['whitelist_keys'] = whitelist_keys
            if len(doubled) > 0:
                k_val['double'] = doubled
                key_info['double'] = doubled

            # add key 'unit' if key is of special case value_unit
            if k_val['input_type'] == 'value_unit':
                k_val['unit'] = None

            # add properties from the general structure
            k_val['displayName'] = node['value'][k]['display_name']
            k_val['required'] = node['value'][k]['mandatory']
            k_val['position'] = k
            k_val['value'] = []
            k_val['desc'] = node['value'][k]['desc']

            # change single_autofill to multi_autofill because all factors can
            # have multiple values
            if k_val['input_type'] == 'single_autofill':
                k_val['input_type'] = 'multi_autofill'

            if k_val['input_type'] == 'restricted_short_text':

                # default
                k_val['restriction'] = {'regex': ' ', 'max_length': 10}

                if 'special_case' in node['value'][k] and 'restriction' in \
                        node['value'][k]['special_case']:
                    if 'regex' in node['value'][k]['special_case'][
                        'restriction']:
                        k_val['restriction']['regex'] = \
                            node['value'][k]['special_case']['restriction'][
                                'regex']
                    if 'max_length' in node['value'][k]['special_case'][
                        'restriction']:
                        k_val['restriction']['max_length'] = \
                            node['value'][k]['special_case']['restriction'][
                                'max_length']

            # add search info if the input is of type single- or multi-autofill
            if k_val['input_type'] == 'multi_autofill':
                k_val['whitelist'] = None
                k_val['search_info'] = {'organism': filled_object['organism'],
                                        'key_name': k}

            # add dictionary with properties of the key to the whitelist
            whitelist.append(k_val)
            if len(key_info) > 0:
                nested_infos[k] = key_info

        # set input type to nested
        input_type = 'nested'
        if len(nested_infos) == 0:
            nested_infos = None

    # value is not a dictionary or special_case: merge or value_unit
    else:

        # read and parse whitelist
        whitelist, whitelist_type, input_type, headers, w_keys, double = \
            whitelist_parsing.parse_whitelist(key, node,
                                              filled_object, whitelist_object=read_in_whitelists)

    # factor takes a list as value -> can occur multiple times in one condition
    if node['list']:

        # values are single values (no dictionaries)
        if not isinstance(node['value'], dict):

            # change single_autofill to multi_autofill because all factors can
            # have multiple values
            if input_type == 'single_autofill':
                input_type = 'multi_autofill'

    return whitelist, whitelist_type, input_type, headers, w_keys, double, nested_infos


def get_conditions(factors, organism_name, key_yaml, read_in_whitelists):
    """
    This function creates all combinations of selected experimental factors and
    their values
    :param key_yaml: the read in general structure
    :param organism_name: the selected organism
    :param factors: multiple dictionaries containing the keys 'factor' and
    'values' with their respective values grouped in a list
    e.g. [{'factor': 'gender', 'values': ['male', 'female']},
          {'factor: 'life_stage', 'values': ['child', 'adult']}]
    :return: a dictionary containing the following keys:
             conditions: a condition object containing a list of all conditions
                         saved as dictionaries
             whitelist_object: a dictionary containing all keys that store a
                               whitelist and their whitelist
             organism: the organism that was chosen by the user
    """

    # initialize an empty dictionary to store the values as they are displayed
    # on the website -> used if there are headers or whitelist keys
    # e.g. value 'GFP' is displayed as 'GFP (other)' on the website since the
    # value is stored in a grouped whitelist that was refactored to plain
    real_val = {}

    # iterate over factors
    for i in range(len(factors)):

        # extract the properties of the factor from the general structure
        factor_infos = list(utils.find_keys(key_yaml, factors[i]['factor']))

        # sanity check -> factor was found in general structure and
        # list of values contains only one element
        if len(factor_infos) > 0 and len(factors[i]['values']) == 1:

            # set val to the values specified for the factor
            val = factors[i]['values'][0]

            # value is a dictionary -> e.g. disease
            if isinstance(val, dict):

                # remove keys with value None or empty lists and dictionaries
                val = {k: v for k, v in val.items() if
                       not (type(v) in [list, dict] and len(v) == 0)
                       and v is not None}

                # add an ident key to the value
                # -> defined in general structure or None
                val['ident_key'] = factor_infos[0]['special_case']['group'] if \
                    'special_case' in factor_infos[0] and 'group' in \
                    factor_infos[0]['special_case'] else None

                val['control'] = factor_infos[0]['special_case']['control'] if \
                    'special_case' in factor_infos[0] and 'control' in \
                    factor_infos[0]['special_case'] else None

                if 'nested_infos' in factors[i]:
                    for val_key in factors[i]['nested_infos']:
                        if val_key in val:
                            for v in range(len(val[val_key])):
                                full_value = copy.deepcopy(val[val_key][v])
                                whitelist_key = None
                                if 'whitelist_keys' in factors[i]['nested_infos'][
                                    val_key]:
                                    for w_key in \
                                    factors[i]['nested_infos'][val_key][
                                        'whitelist_keys']:
                                        if val[val_key][v].endswith(f'({w_key})'):
                                            val[val_key][v] = val[val_key][
                                                v].rstrip(f'({w_key})').strip()
                                            whitelist_key = w_key
                                            break
                                if 'headers' in factors[i]['nested_infos'][
                                    val_key]:
                                    if isinstance(factors[i]['nested_infos'][val_key][
                                            'headers'], dict):
                                        if whitelist_key is not None and whitelist_key in factors[i]['nested_infos'][val_key]['headers']:
                                            str_value = wi_utils.parse_headers(
                                            factors[i]['nested_infos'][val_key][
                                                'headers'][whitelist_key], val[val_key][v],
                                            mode='str')
                                            val[val_key][v] = f'{"{"}' \
                                                          f'{str_value}{"}"}'
                                    else:
                                        str_value = wi_utils.parse_headers(
                                            factors[i]['nested_infos'][val_key][
                                                'headers'], val[val_key][v],
                                            mode='str')
                                        val[val_key][v] = f'{"{"}' \
                                                          f'{str_value}{"}"}'
                                real_val[val[val_key][v]] = full_value

                # generate combinations of the values of the dictionary for the
                # conditions and overwrite the values with them
                factors[i]['values'] = utils.get_combis(
                    val, factors[i]['factor'],
                    {'organism': organism_name.split(' ')[0]}, key_yaml, read_in_whitelists=read_in_whitelists)

        elif factor_infos[0]['list']:

            new_values = []
            for j in range(len(factors[i]['values'])):
                single_val = copy.deepcopy(factors[i]['values'][j])
                single_val_key = None
                if 'whitelist_keys' in factors[i]:
                    for w_key in factors[i]['whitelist_keys']:
                        if single_val.endswith(f'({w_key})'):
                            single_val = single_val.rstrip(f'({w_key})').strip()
                            single_val_key = w_key
                            break
                if 'headers' in factors[i]:
                    if isinstance(factors[i]['headers'], dict):
                        if single_val_key is not None:
                            if single_val_key in factors[i]['headers']:
                                single_val = wi_utils.parse_headers(
                                    factors[i]['headers'][single_val_key],
                                    single_val, mode='dict')
                    else:
                        single_val = wi_utils.parse_headers(
                            factors[i]['headers'],
                            single_val, mode='dict')
                if isinstance(single_val, dict):
                    val = "|".join([f'{key}:"{single_val[key]}"' for key in single_val])
                    val = f'{factors[i]["factor"]}:{"{"}{val}{"}"}'
                    real_val[val] = factors[i]['values'][j]
                new_values.append(single_val)
            factors[i]['values'] = utils.get_combis(new_values,
                                                    factors[i]['factor'], {
                                                        'organism':
                                                            organism_name.split(
                                                                ' ')[0]},
                                                    key_yaml, read_in_whitelists=read_in_whitelists)

        # iterate over all values
        for j in range(len(factors[i]['values'])):

            # factor contains whitelist keys
            if 'whitelist_keys' in factors[i]:

                # save the original value as full_value
                full_value = copy.deepcopy(factors[i]['values'][j])

                # save the headers if they are specified for the factor else
                # None
                headers = factors[i]['headers'] if 'headers' in factors[i] \
                    else None

                # rewrite the values by removing the whitelist keys and split
                # them according to the headers if headers were defined
                factors[i]['values'][j] = wi_utils.parse_whitelist_keys(
                    factors[i]['whitelist_keys'], factors[i]['values'][j],
                    headers, mode='str')

                # headers were defined and the whitelist key of the value is
                # defined within the headers
                if headers is not None and full_value.split('(')[-1].replace(
                        ')', '') in headers:
                    # rewrite the value to '<factor>:{<values>}'
                    factors[i]['values'][j] = f'{factors[i]["factor"]}:{"{"}' \
                                              f'{factors[i]["values"][j]}{"}"}'

                # save the original value in real_val with the new value as key
                if factors[i]['values'][j] not in real_val:
                    real_val[factors[i]['values'][j]] = full_value

            # factor contains headers but not whitelist keys
            elif 'headers' in factors[i]:

                # save the original value
                full_value = copy.deepcopy(factors[i]['values'][j])

                # split the value according to the header and save them as a
                # string
                str_value = wi_utils.parse_headers(
                    factors[i]['headers'], factors[i]['values'][j], mode='str')

                # # rewrite the value to '<factor>:{<values>}'
                factors[i]['values'][j] = f'{factors[i]["factor"]}:{"{"}' \
                                          f'{str_value}{"}"}'

                # save the original value in real_val with the new value as key
                if factors[i]['values'][j] not in real_val:
                    real_val[factors[i]['values'][j]] = full_value

    # generate the conditions
    conditions = utils.get_condition_combinations(factors)

    # extract the properties of 'sample' from the general structure
    sample = list(utils.find_keys(key_yaml, 'samples'))

    # initialize a condition- and a whitelist object
    condition_object = []
    whitelist_object = {}

    # sanity check -> sample was found in the general structure
    if len(sample) > 0:

        # create an emtpy wi_object and a whitelist object from the sample
        # structure
        sample, whitelist_object = yto.parse_empty(
            sample[0],
            'experimental_setting:conditions:biological_replicates:samples',
            key_yaml, {'organism': organism_name}, read_in_whitelists, get_whitelist_object=True)

        # overwrite sample with its input fields
        sample_desc = sample['desc']
        sample = sample['input_fields']

        # iterate over conditions
        for cond in conditions:

            # generate a sample name from the condition
            sample_name = utils.get_short_name(cond,
                                               {'organism': organism_name},
                                               key_yaml, read_in_whitelists=read_in_whitelists)

            # split the condition into key-value pairs
            split_condition = utils.split_cond(cond)
            search_condition = []

            # TODO: remove list (can there be a list?) or own function for items
            for elem in split_condition:
                if isinstance(elem[1], dict):
                    val = "|".join([f'{k}:"{elem[1][k]}"' for k in elem[1]])
                    val = f'{elem[0]}:{"{"}{val}{"}"}'
                    if val in real_val:
                        search_condition.append(
                            f'{elem[0]}:"{real_val[val]}"'.lower())
                    else:
                        for key in elem[1]:
                            if isinstance(elem[1][key], dict):
                                val = "|".join(
                                    [f'{k}:"{elem[1][key][k]}"' for k in
                                     elem[1][key]])
                                val = f'{key}:{"{"}{val}{"}"}'
                                if val in real_val:
                                    search_condition.append(
                                        f'{elem[0]}:"{real_val[val]}"'.lower())
                                else:
                                    # TODO: real/val or own function to get values from deeper dictionaries
                                    for sub_key in elem[1][key]:
                                        search_condition.append(f'{elem[0]}:"{elem[1][key][sub_key]}"'.lower())
                            else:
                                search_condition.append(
                                    f'{elem[0]}:"{elem[1][key]}"'.lower())
                elif isinstance(elem[1], list):
                    for item in elem[1]:
                        search_condition.append(f'{elem[0]}:"{item}"'.lower())
                else:
                    if elem[1] in real_val:
                        search_condition.append(
                            f'{elem[0]}:"{real_val[elem[1]]}"'.lower())
                    else:
                        search_condition.append(
                            f'{elem[0]}:"{elem[1]}"'.lower())
            # call functions to fill the samples with the values from the
            # condition
            filled_sample = get_samples(split_condition, copy.deepcopy(sample),
                                        real_val, key_yaml, sample_name,
                                        organism_name, read_in_whitelists)

            # rewrite condition into title and string to readd deleted
            # conditions
            title, readd = get_condition_title(split_condition)

            # save the condition as a dictionary with the filled sample as
            # input fields
            d = {'correct_value': cond,
                 'title': title,
                 'search': search_condition,
                 'readd': readd,
                 'position': 'experimental_setting:condition',
                 'list': True, 'mandatory': True, 'list_value': [],
                 'input_disabled': False, 'desc': sample_desc,
                 'input_fields': copy.deepcopy(filled_sample)}

            # add the dictionary for the condition to the condition object
            condition_object.append(d)

    return {'conditions': condition_object,
            'whitelist_object': whitelist_object, 'organism': organism_name}


def get_condition_title(split_condition):
    """
    This function rewrites a condition into a table to be displayed as a title
    on the website. It also creates a string that is shown when removed
    conditions have to be re-added
    :param split_condition: a list containing the factors and values the
                            condition consists of
    :return: html: the title as a table in html
             readd: the string shown when re-adding deleted conditions
    """

    # initialize the html string with the table header and style
    html = '<table class="table_style_condition_title"><tbody>'

    # initialize readd as an empty string
    readd = ''

    # iterate over the factors and values
    for i in range(len(split_condition)):

        # start the row of the table with the style depending on the index of
        # the element that is written into the row (the style creates a
        # horizontal line below the row and includes padding)
        if len(split_condition) > 1 and i < len(split_condition) - 1:
            html += '<tr class="tr_style_condition_title">'
        else:
            html += '<tr>'

        # add the factor to readd
        readd += f'{split_condition[i][0]}:\n'

        # add the factor to the table as a data cell with the style of a value
        # (padding)
        html += f'<td class="td_style_condition_title_value">' \
                f'{split_condition[i][0]}:</td>'

        vals, readd = get_title_value(split_condition[i][1], readd)

        # add the string with the values to the table as a data cell
        html += f'<td class="td_style_condition_title_value">{vals}' \
                f'</td></tr>'

        # if the condition consists of multiple factors than write a '---'
        # between them in readd
        if len(split_condition) > 1 and i < len(split_condition) - 1:
            readd += '---\n'

    # close the html table
    html += f'</tbody></table>'

    return html, readd


def get_title_value(cond_value, readd):
    # value is a dictionary
    if isinstance(cond_value, dict):

        # initialize an empty string to save the values to
        vals = ''

        # iterate over the keys
        for key in cond_value:

            value, readd = get_title_value(cond_value[key], readd)

            # add the value to the string with a <br> if it is not in the
            # last row
            if key != list(cond_value.keys())[-1]:
                vals += f'{value}<br>'
            else:
                vals += f'{value}'

    # value is not a dictionary
    else:

        # replace more than 3 zeros with 3 dots to shorten the value
        vals = f'"{re.sub("0000(0)*", "...", cond_value)}"'

        # add the value to readd
        readd += f'{vals}\n'

    return vals, readd


def get_samples(split_condition, sample, real_val, key_yaml, sample_name,
                organism_name, read_in_whitelists, is_factor=True):
    """
    This function created a pre-filled object with the structure of the samples
    to be displayed in the web interface
    :param is_factor: a boolean defining if the input field is an experimental
                      factor
    :param organism_name: the name of the selected organism
    :param sample_name: the identifier of the sample build from the condition
    :param key_yaml: the read in general structure
    :param real_val: a dictionary containing the values containing headers and
                     whitelist keys as they should be displayed on the website
    :param split_condition: the condition the sample is created for
    :param sample: the empty structure of the sample
    :return: sample: the pre-filled sample
    """
    # save all factors in a list
    factors = [cond[0] for cond in split_condition]

    # iterate over samples
    for i in range(len(sample)):

        # input field: sample_name
        if sample[i]['position'].endswith('samples:sample_name'):

            # TODO: improve display
            # add whitespaces to sample name to enable line breaks on the
            # website
            sample[i]['value'] = sample_name \
                .replace(':', ': ') \
                .replace('|', '| ') \
                .replace('#', '# ') \
                .replace('-', ' - ') \
                .replace('+', ' + ')

            # save the unchanged sample name as 'correct_value'
            sample[i]['correct_value'] = sample_name.split('_')[0]

        # elif sample[i]['position'].endswith('samples:sequencer_name'):
        #    sample[i]['value'] = sequencer_name
        #    sample[i]['input_disabled'] = True

        # input field is a factor
        elif sample[i]['position'].split(':')[-1] in factors:

            # iterate over factors in condition
            for c in split_condition:

                # input field of current factor
                if sample[i]['position'].split(':')[-1] == c[0]:
                    # extract properties of the factor from the general
                    # structure
                    info = list(utils.find_keys(key_yaml, c[0]))

                    # sanity check -> factor was found in general structure
                    if len(info) > 0:

                        filled_value = ''
                        # special case value_unit
                        if 'special_case' in info[0] and 'value_unit' \
                                in info[0]['special_case']:

                            if isinstance(c[1], dict):
                                value_unit = c[1]
                            else:
                                # split the value into value and unit
                                value_unit = utils.split_value_unit(c[1])

                            # save the value and unit in the sample
                            sample[i]['value'] = value_unit['value']
                            sample[i]['value_unit'] = value_unit['unit']
                            if is_factor:
                                sample[i]['input_disabled'] = True

                        # value is a dictionary
                        elif isinstance(c[1], dict):

                            # rewrite the value into a string
                            val = "|".join(
                                [f'{key}:"{c[1][key]}"' for key in c[1]])
                            val = f'{c[0]}:{"{"}{val}{"}"}'
                            # save the value from the dictionary real_val if
                            # real_val contains the current value as key

                            if val in real_val:
                                filled_value = real_val[val]

                            # value is not in real_val
                            else:

                                # TODO: rework with real_val

                                if 'input_fields' in sample[i]:
                                    # call this function on the keys of the
                                    # value in order to fill them
                                    filled_value = get_samples(
                                        [(x, c[1][x]) for x in c[1]  # if not (
                                         # c[1] == 'technical_replicates'
                                         # and x == 'sample_name')
                                         ],
                                        copy.deepcopy(
                                            sample[i]['input_fields']),
                                        info, key_yaml, sample_name,
                                        organism_name, read_in_whitelists, is_factor=is_factor)
                                # TODO: als real_val? automate
                                elif c[0] == 'enrichment_type':
                                    connected_value = []
                                    for key in c[1]:
                                        connected_value.append(c[1][key])
                                    connected_value = ' '.join(connected_value)
                                    filled_value = f'{connected_value} (proteins)'
                                    sample[i]['headers'] = {'proteins': 'gene_name ensembl_id'}
                                elif 'headers' in sample[i]:
                                    headers = [x for x in c[1]]
                                    w_key = None
                                    if isinstance(sample[i]['headers'], dict):

                                        for k in sample[i]['headers']:
                                            if sorted(
                                                    sample[i]['headers'][
                                                        k].split(' ')) == \
                                                    sorted(headers):
                                                w_key = k
                                                break
                                    else:
                                        if sorted(headers) != sorted(
                                                sample[i]['headers'].split(
                                                    ' ')):
                                            headers = None

                                    if headers is not None:
                                        filled_value = ''
                                        for header in headers:
                                            filled_value = filled_value + ' ' \
                                                           + c[1][header]
                                        filled_value = filled_value.lstrip(
                                            ' ').rstrip(' ')

                                    if filled_value is not None and w_key is \
                                            not None and \
                                            sample[i]['whitelist_type'] == \
                                            'plain_group':
                                        filled_value = f'{filled_value} ' \
                                                       f'({w_key})'
                            # save the filled value in 'list_value' if the
                            # input field takes a list
                            if sample[i]['list']:
                                sample[i]['list_value'].append(filled_value)

                            # save the filled value in 'input_fields' if the
                            # input field takes a dictionary
                            elif 'input_fields' in sample[i]:
                                sample[i]['input_fields'] = filled_value

                            # save the filled value in 'value'
                            else:
                                sample[i]['value'] = filled_value

                        # value is not a dictionary
                        else:
                            # save the value from the dictionary real_val if it
                            # contains the current value
                            if c[0] == 'enrichment_type' and 'headers' in sample[i] and 'proteins' in sample[i]['headers']:
                                if not c[1].endswith('(proteins)'):
                                    filled_value = f'{c[1]} (proteins)'
                                else:
                                    filled_value = c[1]

                            elif c[1] in real_val:
                                filled_value = real_val[c[1]]

                            # save the current value
                            else:

                                if sample[i]['whitelist_type'] == \
                                        'plain_group' and 'whitelist_keys' \
                                        in sample[i]:
                                    w = utils.get_whitelist(
                                        sample[i]['position'].split(':')[-1],
                                        {'organism': organism_name}, whitelist_object=read_in_whitelists)

                                    for key in sample[i]['whitelist_keys']:
                                        filled_value = c[1]
                                        if f'{c[1]} ({key})' in w['whitelist']:
                                            filled_value = f'{c[1]} ({key})'
                                            break
                                else:
                                    if 'input_type' in info[0] and info[0][
                                        'input_type'] == 'number':
                                        filled_value = int(c[1])
                                    else:
                                        filled_value = c[1]
                            # save the filled value in 'list_value' if the
                            # input field takes a list
                            if sample[i]['list']:
                                sample[i]['list_value'].append(filled_value)

                            # save the filled value in 'value'
                            else:
                                sample[i]['value'] = filled_value

                        # input field is of type single_autofill
                        if 'input_type' in sample[i] and \
                                sample[i]['input_type'] in [
                            'single_autofill', 'restricted_short_text']:
                            # initialize the key 'list_value' and move the
                            # value under the key 'value' to the key
                            # 'list_value'
                            sample[i]['list_value'] = [] if \
                                sample[i]['value'] is None \
                                else [sample[i]['value']]
                            sample[i]['value'] = None

                        if is_factor:

                            if 'list' in info[0] and info[0]['list']:
                                sample[i]['delete_disabled'] = True
                                sample[i]['fixed_length'] = len(
                                    sample[i]['list_value'])
                            else:
                                # disable the input for the filled input field
                                sample[i]['input_disabled'] = True

                        if 'input_type' in info[0] and info[0][
                            'input_type'] == 'bool':
                            if sample[i]['value']:
                                sample[i]['value'] = 'True'
                            else:
                                sample[i]['value'] = 'False'

    return sample



from dash import Dash, html, dcc, callback, Output, Input
import plotly.express as px
import plotly.graph_objects as go
import plotly as plt
import pandas as pd
import argparse
import pathlib
from src import utils
from src.heatmap.dash_utils import *
import numpy as np
from src.heatmap.dash_utils import get_data
import os
import base64
import io
from PIL import Image

parser = argparse.ArgumentParser('Metadata Heatmap')
parser.add_argument('-p', '--path', type=pathlib.Path, required=True, help='The path of the metadata file')

args = parser.parse_args()
input_file = utils.read_in_yaml(args.path)
key_yaml = utils.read_in_yaml('keys.yaml')

settings, experimental_factors, organisms, max_vals, options_pretty, annotated_dict = get_data(input_file, key_yaml)


app = Dash()

# Requires Dash 2.17.0 or later
app.layout = [
    dcc.Tabs(id="tabs-example-graph", value=f'tab-{list(settings.keys())[0]}', 
             children=[dcc.Tab(label=f'Setting {x}', value=f'tab-{x}') for x in settings]),
                       dcc.Checklist(['Show Empty'],[], id='checklist-selection'),
                       dcc.Graph(id='graph-content')
                       ]

@callback(
    Output('graph-content', 'figure'),
    Input('tabs-example-graph', 'value'),
    Input('checklist-selection', 'value')
)
def update_graph(value, empty):
    value = value.replace('tab-','')
    options = [key.replace('_num', '') for key in settings[value] if key.endswith('_num')]
    sorter = experimental_factors[value] + [o for o in options if o not in experimental_factors[value]]

    if not empty:
        df_empty = settings[value].dropna(axis=1, how='all')
        sorter = [x for x in sorter if x in df_empty.columns]
    df = [settings[value][f'{key}_num'] for key in sorter]   

    annotated = [annotated_dict[value][key] for key in sorter if key in annotated_dict[value]]
    label_text = []
    for key in sorter:
        if key in annotated_dict[value] and key in experimental_factors[value]:
            label_text.append(annotated_dict[value][key])
        else:
            label_text.append('')

    option_text = []

    for option in sorter:
        if option in experimental_factors[value]:
            option_text.append(color("red", f'<b>{options_pretty[option]}</b>'))
        else:
            option_text.append(color("black", options_pretty[option]))

    plotly_colors = [x for x in px.colors.qualitative.G10 if x != "#DC3912"]
    print(plotly_colors)
    colors = [[0, 'white']]
    for i in range(1, max_vals[value]+1):
        colors.append([i*1/(max_vals[value]), plotly_colors[i-1]])
        if i*1/(max_vals[value]) != 1:
            colors.append([((i*1)/(max_vals[value]))+0.001, 'white'])

    heatmap = [go.Heatmap(
                z=df,
                zmin=0,
                zmax=max_vals[value],
                x=[settings[value]['condition_index'],settings[value]['sample_index']],
                y=sorter,
                showscale=False,
                customdata=annotated,
                text=label_text,
                texttemplate="%{text}",
                hovertemplate = "%{customdata}",
                hoverongaps = False,
                colorscale = colors,
                ),
            ]
        
    condition_labels = {}
    for i in range(len(settings[value]['condition_name'])):
        if settings[value]['condition_index'][i] not in condition_labels:
            cond = settings[value]['condition_name'][i]
            splitted = utils.split_cond(cond)
            cond_dict = {}
            for elem in splitted:
                if isinstance(elem[1], dict):
                    vals = [elem[1][k] for k in elem[1]]
                else:
                    vals =  [elem[1]]
                if elem[0] in cond_dict:
                    cond_dict[elem[0]] += vals
                else:
                    cond_dict[elem[0]] = vals
            condition_labels[settings[value]['condition_index'][i]]= cond_dict

    data_input = heatmap
        
    my_cell_width = 150
    top_margin = 100
    bottom_margin = 0
    left_margin = 200
    right_margin = 200
    my_height = 50*len(sorter)
    my_width = my_cell_width*len(settings[value]["sample_index"])

    organism_path = os.path.join(os.path.dirname(__file__), 'images', f'{organisms[value]}.png')
    images = None

    if os.path.isfile(organism_path):
        plotly_logo = base64.b64encode(open(organism_path, 'rb').read())
        imgdata = base64.b64decode(plotly_logo)
        im = Image.open(io.BytesIO(imgdata))
        im_width, im_height = im.size
        y_side = top_margin
        my_ysize=y_side/my_height
        x_side = y_side*im_width / im_height
        my_xsize = x_side/my_width

        left_margin = max(200, x_side)

        images = [dict(
            source='data:image/png;base64,{}'.format(plotly_logo.decode()),
            xref="paper", yref="paper",
            x=0, y=1,
            sizey=my_ysize, sizex=my_xsize,
            xanchor="right", yanchor="bottom"
        )]


    layout = go.Layout(
        images = images,
        height=my_height + top_margin + bottom_margin,
        width=my_width + left_margin + right_margin,
        margin=dict(l=left_margin, r=right_margin, t=top_margin, b=bottom_margin),
        autosize=False,
        title=dict(
            text=f'<b>Setting {value}</b>', 
            font=dict(size=15), 
            automargin=False,
            yref='container',
            x=(left_margin+(0.1*150))/(my_width+left_margin+right_margin),
            y=1,
            xanchor="left", yanchor="top",
            subtitle=dict(
                text=f'Organism: {organisms[value]}',
                font=dict(size=14, lineposition='under'),
                ),
            ),
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)',
        xaxis=dict(
            side="top",
            showline=True,
            gridcolor="lightgrey",
            automargin=False
        ),
        yaxis=dict(
            tickmode="array",
            ticktext=option_text, 
            tickvals=sorter,
            autorange="reversed",
            tickson="boundaries",
            showline=True,
            gridcolor="lightgrey",
            automargin=False
        ),
        legend=dict(
            #title=f"Organism: {organisms[value]}",
            orientation="h",
            x=0, y=0
        )
    )   
    fig = go.Figure(data=data_input, layout=layout) 

    for i in range(len(sorter)):
        fig.add_hline(i, line_dash="dash", line_color='lightgrey',layer='below')
        fig.add_hline(i-0.5, line_width=0.5) 
    fig.add_hline(len(sorter)-0.5, line_width=0.5)     

    for i in range(len(settings[value]['sample_index'])+1):
        fig.add_vline(i-0.5, line_width=0.5)
    
    fig.add_hrect(y0=-0.5, y1=len(experimental_factors[value])-0.5, line=dict(color="red", width=5), layer='above')
        
    return fig

def color(color, text):
    return f"<span style='color:{str(color)}'> {str(text)} </span>"

if __name__ == '__main__':

    app.run(debug=True)


import pandas as pd
from src import utils
import numpy as np
import re

def get_data(path, keys_yaml, mode='samples', drop_defaults=False, transpose=False):

    setting_dict = {}
    annotated_dict = {}
    experimental_factors = {}
    organisms = {}
    option_types = {}
    option_pretty = {}
    max_vals = {}
    max_anno = {}
    no_samples = {}

    yaml = path
    keys = keys_yaml
    settings = list(utils.find_keys(yaml, 'experimental_setting'))

    if len(settings) > 0:

        options = [x for x in list(utils.find_keys(keys, 'samples'))[0]['value'].keys()if x not in ['sample_name']]

        for setting_index, setting in enumerate(settings[0]):
            
            df_dict = {}
            max_annotation = {}

            try:
                setting_id = setting['setting_id']
            except KeyError:
                setting_id = f'setting_{setting_index}'
            
            try:
                organisms[setting_id] = setting['organism']['organism_name']
            except KeyError:
                organisms[setting_id] = None

            try:
                experimental_factors[setting_id] = [x['factor'] for x in setting['experimental_factors']]
            except KeyError:
                experimental_factors[setting_id] = None

            conditions = list(utils.find_keys(setting, 'conditions'))
            
            if len(conditions) > 0:
                
                conditions = conditions[0]
                condition_names = []
                condition_index = []
                sample_index = []
                sample_label = []
                sample_label_height = []
                condition_label_height = []
                condition_label = []
                idx = []
                option_dict = {}
                annotated = {}
                no_samples[setting_id] = {}
                
                j=1
                
                for cond_index, cond in enumerate(conditions):
                
                    condition_name = cond['condition_name']
                    c_index = f'<b>Condition {cond_index+1}</b>'

                    splitted_condition = utils.split_cond(condition_name)
                    missing_dict = {}
                    
                    if mode == 'samples':
                        samples = list(utils.find_keys(cond, 'samples'))
                        
                        if len(samples) > 0:
                            samples = samples[0]
                    
                            for samp_index, sample in enumerate(samples):
                                
                                condition_names.append(condition_name)
                                condition_index.append(c_index)
                                
                                sample_index.append(f'Sample {samp_index+1}')

                                idx.append(j)
                                j+=1

                                for key in options:
                                    
                                    if key not in option_dict:
                                        option_dict[key] = []
                                        annotated[key] = []
                                    
                                    if key in sample:    
                                        if key == 'technical_replicates':
                                            try:
                                                option_dict[key].append(sample[key]['count'])
                                                annotated[key].append(str(sample[key]['count']))
                                            except KeyError:
                                                option_dict[key].append(None)
                                                annotated[key].append('')
                                            
                                            try:
                                                max_width = 20
                                                label=sample[key]["filenames"][0].split('__')[-2].replace("_", "-")
                                                if len(label)+len(f'_{samp_index+1}')>=max_width:
                                                    label_height = 1
                                                    splitted = label.split('-')
                                                    new_label = f'{splitted[0]}-'
                                                    row_len = len(new_label)
                                                    for i in range(1, len(splitted)):
                                                        elem = splitted[i] + '-' if i<len(splitted)-1 else splitted[i] + f'_{samp_index+1}'
                                                        if row_len+len(elem) >= max_width:
                                                            label_height += 1
                                                            new_label += '<br>'
                                                            row_len = len(elem)
                                                        else:
                                                            row_len += len(elem)
                                                        new_label += f'{elem}'
                                                    label = new_label
                                                    sample_label_height.append(label_height)
                                                    condition_label_height.append(label_height)
                                                else:
                                                    label = f'{label}_{samp_index+1}'
                                                    sample_label_height.append(1)
                                                    condition_label_height.append(1)
                                                sample_label.append(label)
                                                condition_label.append(f'<b>{label.replace(f"_{samp_index+1}", "")}</b>')
                                            except (KeyError, IndexError):
                                                sample_label.append(f'Sample {samp_index+1}')
                                                condition_label.append(f'<b>Condition {cond_index+1}</b>')
                                        else:

                                            if isinstance(sample[key], int) or isinstance(sample[key], float):
                                                option_dict[key].append(sample[key])
                                                
                                            elif (isinstance(sample[key], dict) and 'value' in sample[key] and 'unit' in sample[key]):
                                                option_dict[key].append(sample[key])
                                                
                                            else:
                                                option_dict[key].append(str(sample[key]))
                                                
                                            key_structure = list(utils.find_keys(keys, key))
                                            group_key = None
                                            if len(key_structure) > 0:
                                                try:
                                                    group_key = list(key_structure[0]['special_case']['control'].keys())[0]
                                                except:
                                                    group_key = None
                                            annotation = annotate(sample[key], group_key)
                                            if transpose:
                                                new_annotation = []
                                                single_val = ''
                                                for i in range(len(annotation)):
                                                    if len(single_val) + (1 if i>0 else 0) + len(annotation[i]) < 20:
                                                        single_val += f'{" " if i > 0 else ""}{annotation[i]}'
                                                    else:
                                                        new_annotation.append(single_val)
                                                        single_val = annotation[i]
                                                if single_val != '':
                                                    new_annotation.append(single_val)
                                                annotation = new_annotation
                                            if key in max_annotation:
                                                max_annotation[key] = max(max_annotation[key], len(annotation))
                                            else:
                                                max_annotation[key] = len(annotation)
                                            annotated[key].append('<br>'.join(annotation))
                                    else:
                                        option_dict[key].append(None)
                                        annotated[key].append('')
                        else:
                            for elem in splitted_condition:
                                if elem[0] in missing_dict:
                                    missing_dict[elem[0]].append(elem[1])
                                else:
                                    missing_dict[elem[0]] = [elem[1]]

                            no_samples[setting_id][c_index] = missing_dict
                    
                        df_dict = {'condition_name': condition_names,
                                'condition_index': condition_index,
                                'condition_label': condition_label,
                                'condition_label_height': condition_label_height,
                                'sample_index': sample_index,
                                'sample_label': sample_label,
                                'sample_label_height': sample_label_height,
                                'index': idx}
                    elif mode=='conditions':
                        
                        try:
                            sample_count = cond['biological_replicates']['count']
                        except KeyError:
                            sample_count = 0

                        if 'sample_count' not in option_dict:
                            option_dict['sample_count'] = []
                        option_dict['sample_count'].append(sample_count)

                        condition_names.append(condition_name)
                        
                        short = utils.get_short_name(condition_name, {}, key_yaml=keys_yaml)
                        
                        c_label=[]
                        for elem in short.split('-'):
                            for part in elem.split('+'):
                                c_label.append(part.split('.')[-1])

                        break_label = c_label[0] + '-'
                        row_len = len(break_label)
                        label_height = 1

                        for i in range(1,len(c_label)):
                            elem = c_label[i] + '-' if i < len(c_label)-1 else c_label[i]
                            if row_len + len(elem) >= 20:
                                break_label += '<br>'
                                label_height +=1
                                row_len = len(elem)
                            else:
                                row_len += len(elem)
                            break_label += elem
                        c_label = break_label

                        condition_label_height.append(label_height)
                        condition_label.append(c_label)
                        condition_index.append(c_index)

                        for option in options:
                            
                            if option not in option_dict:
                                    option_dict[option] = []
                                    annotated[option] = []

                            if option in [x[0] for x in splitted_condition]:

                                my_index = [x[0] for x in splitted_condition].index(option)

                                key = splitted_condition[my_index][0]
                                
                                key_structure = list(utils.find_keys(keys_yaml, key))
                                input_type = None
                                if len(key_structure) > 0 and 'special_case' in key_structure[0] and 'value_unit' in key_structure[0]['special_case']:
                                    input_type = 'value_unit'

                            
                                if input_type == 'value_unit':
                                    value_unit = utils.split_value_unit(splitted_condition[my_index][1])
                                    option_dict[key].append(value_unit)
                                
                                elif isinstance(splitted_condition[my_index][1], int) or isinstance(splitted_condition[my_index][1], float):
                                    option_dict[key].append(splitted_condition[my_index][1])
                                                
                                else:
                                    option_dict[key].append(str(splitted_condition[my_index][1]))
                                
                                key_structure = list(utils.find_keys(keys, key))
                                group_key = None

                                if len(key_structure) > 0:
                                    try:
                                        group_key = list(key_structure[0]['special_case']['control'].keys())[0]
                                    except:
                                        group_key = None
                                    annotation = annotate(splitted_condition[my_index][1], group_key)
                                    if transpose:
                                        new_annotation = []
                                        single_val = ''
                                        for i in range(len(annotation)):
                                            if len(single_val) + (1 if i>0 else 0) + len(annotation[i]) < 20:
                                                single_val += f'{" " if i > 0 else ""}{annotation[i]}'
                                            else:
                                                new_annotation.append(single_val)
                                                single_val = annotation[i]
                                        if single_val != '':
                                            new_annotation.append(single_val)
                                        annotation = new_annotation
                                    if key in max_annotation:
                                        max_annotation[key] = max(max_annotation[key], len(annotation))
                                    else:
                                        max_annotation[key] = len(annotation)
                                    annotated[key].append('<br>'.join(annotation))
                                    
                            else:
                                option_dict[option].append(None)
                                annotated[option].append('')
                            
                            df_dict = {'condition_name': condition_names,
                                        'condition_index': condition_index,
                                        'condition_label': condition_label,
                                        'condition_label_height': condition_label_height
                                        } 
                                            
                    
                for option in option_dict:

                    option_structure = list(utils.find_keys(keys, option))
                        
                    if len(option_structure) > 0:
                        option_structure = option_structure[0]
                            
                        if option not in option_pretty:
                            option_pretty[option] = option.replace('_', ' ').title()
                            
                        if option == 'technical_replicates':
                            option_types[option] = 'number'

                        elif 'input_type' in option_structure:
                            option_types[option] = option_structure['input_type']
                            
                        elif 'special_case' in option_structure and 'value_unit' in option_structure['special_case']:
                            option_types[option] = 'value_unit'
                            
                        else:
                            option_types[option] = 'nested'
                    
                    elif option == 'sample_count':
                        option_pretty[option] = option.replace('_', ' ').title()
                        option_types[option] = 'number'    
                        annotated[option] = [str(x) for x in option_dict[option]]
                try:
                    max_options = max([len(set([str(y) for y in option_dict[x] if y is not None])) for x in option_dict.keys() if option_types[x] not in ['number', 'value_unit']])
                except ValueError:
                    max_options = 0
                
                try:
                    max_number = max([max([y for y in option_dict[x] if y is not None] + [1]) for x in option_dict.keys() if option_types[x] == 'number'])
                except ValueError:
                    max_number = 0
                    
                vu_values = {}
                    
                for option in option_dict:

                    if option_types[option] == 'value_unit':
                            
                        for i in range(len(option_dict[option])):
                                
                            if option_dict[option][i] is not None:
                                    
                                if option_dict[option][i]['unit'] not in vu_values:
                                        vu_values[option_dict[option][i]['unit']] = {}
                                    
                                if 'vals' not in vu_values[option_dict[option][i]['unit']]:
                                    vu_values[option_dict[option][i]['unit']]['vals'] = []
                                    
                                vu_values[option_dict[option][i]['unit']]['vals'].append(option_dict[option][i]['value'])
                    
                min_vu = 0.1
                max_vu = 1
                    
                for unit in vu_values:
                    vu_values[unit]['max'] = max_vu
                    vu_values[unit]['min'] = min_vu
                    min_vu += 1
                    max_vu += 1
                    
                for option in option_dict:
                    
                    option_structure = list(utils.find_keys(keys, option))

                    option_num = []
                    if option_types[option] == 'value_unit':
                        # TODO: sort by unit
                        option_values = sorted([x for x in option_dict[option] if x is not None], key=lambda d: d['unit'])
                    else:
                        option_values = sorted(list(set([x for x in option_dict[option] if x is not None])))
                    
                    keep = True
                    if drop_defaults and len(option_structure) > 0:
                        option_structure = option_structure[0]
                        if 'value' in option_structure and option_structure['value'] is not None:
                            if option == 'technical_replicates':
                                if all([x == 1 for x in option_values]):
                                    keep = False
                            elif all([x == option_structure['value'] for x in option_values]):
                                keep = False
                    if keep:
                        if len(option_values) > 0:
                            if option_types[option] == 'number':
                                num_vals = [normalize(x, 0, max_number, 0.1, 1) for x in option_values]
                            else:
                                num_vals = list(np.linspace(1, len(option_values), len(option_values)))
                            option_values = sorted([str(x) for x in option_values])
                    else:
                        option_dict[option] = [None] * len(option_dict[option])

                    for value in option_dict[option]:
                        if value is None:
                            option_num.append(value)
                        elif option_types[option] == 'value_unit':
                            option_num.append(normalize(value['value'], 0, max(vu_values[value['unit']]['vals']), vu_values[value['unit']]['min'], vu_values[value['unit']]['max']))
                        else:
                            option_num.append(num_vals[option_values.index(str(value))])

                    df_dict[option] = option_dict[option]
                    df_dict[f'{option}_num'] = option_num
                    
                for option in option_dict:
                    if option_types[option] == 'value_unit':
                        for i in range(len(option_dict[option])):
                            if option_dict[option][i] is not None:
                                option_dict[option][i] = str(option_dict[option][i])

                df = pd.DataFrame(df_dict) 
                setting_dict[setting_id] = df
                annotated_dict[setting_id] = annotated
                max_vals[setting_id] = max_options   
                max_anno[setting_id] = max_annotation              
                    
    return setting_dict, experimental_factors, organisms, max_vals, option_pretty, annotated_dict, max_anno, no_samples


def normalize(x, minIn, maxIn, minOut, maxOut):
    if maxIn == minIn:
        return x
    else:
        input = (x-minIn)/(maxIn-minIn) if maxIn != minIn else x
        return minOut + input * (maxOut-minOut)

def annotate(value, group_key=None):
    annotated_value = []
    if isinstance(value, list):
        if all([isinstance(x, dict) for x in value]):
            if group_key is not None:
                group_vals = [x[group_key] for x in value if group_key in x]
                grouped_value = {}
                for val in group_vals:
                    grouped_value[val] = [val]
                    for elem in value:
                        if group_key in elem and elem[group_key] == val:
                            grouped_value[val] += [x for x in annotate(elem) if x not in grouped_value[val]]
                for key in list(grouped_value.keys()):
                    annotated_value += [x for x in grouped_value[key]]
            else:
                for elem in value:
                    annotated_value += [x for x in annotate(elem) if x not in annotated_value]
            keys = []
            for elem in [list(x.keys()) for x in value]:
                keys += elem
            keys = list(set(keys))

            for key in keys:
                key_vals = []
                for elem in value:
                    if key in elem:
                        key_vals += [x for x in annotate(elem[key]) if x not in key_vals]
                annotated_value += [x for x in key_vals if x not in annotated_value]
        else:
            for elem in value:
                annotated_value += [x for x in annotate(elem) if x not in annotated_value]
    elif isinstance(value, dict):
        if 'value' in value and 'unit' in value:
            annotated_value += [x for x in [f'{value["value"]}{value["unit"]}'] if x not in annotated_value]
        else:
            for key in value:
                if key != 'ensembl_id':
                    annotated_value += [x for x in annotate(value[key]) if x not in annotated_value]
    else:
        annotated_value += [x for x in [re.sub("0000(0)*", "...", str(value))] if x not in annotated_value]
    
    return annotated_value


import copy

from src import utils


class Autogenerate:

    def __init__(self, gen, position):
        """
        initialize class
        :param gen: an object generated from class Generate
        :param position: the position of the key
        """
        self.gen = gen
        self.position = position

    def get_id(self):
        """
        returns the project ID
        :return: project_id from class Generate
        """
        return self.gen.project_id

    def get_setting_id(self):
        """
        creates an ID for the experimental setting
        :return: setting_id (exp1, exp2, ...)
        """

        # search for setting IDs in result_dict
        setting_ids = self.gen.setting_ids

        # find the setting_id with the highest index and add 1
        if len(setting_ids) > 0:
            max_id = 0
            for s_id in setting_ids:
                max_id = max(max_id, int(s_id.replace('exp', '')))
            setting_id = f'exp{max_id+1}'

        # set setting_id to exp1 if it is the first setting
        else:
            setting_id = 'exp1'

        self.gen.setting_ids.append(setting_id)

        return setting_id

    def get_experimental_factors(self):
        """
        request the user to specify the experimental factors
        :return: experimental_factors: list of dictionaries
                 -> [{factor: <fac>, values: [<val>,...]}, ...]
        """


        factor_list = utils.get_whitelist('factor', self.gen.result_dict)['whitelist']
        plain_list = []
        exp_info = list(utils.find_keys(self.gen.key_yaml, 'experimental_factors'))[0]
        factor_info = \
        list(utils.find_keys(self.gen.key_yaml, 'factor'))[0]
        for fk in factor_list:
            plain_list += factor_list[fk]

        print(f'\n---\n{exp_info["factor_desc"]}')
        print(
            f'\nPlease select the analyzed experimental factors '
            f'(1-{len(plain_list)}) divided by comma:\n')

        used_factors = self.gen.get_input_list(factor_info, 'factor')
        experimental_factors = []

        for fac in used_factors:

            fac_node = list(utils.find_keys(self.gen.key_yaml, fac))[0]
            fac_dict = {}
            self.gen.parse_lists(fac_node, [fac], 2, fac_dict, is_factor=True)

            experimental_factors.append({'factor': fac, 'values': fac_dict[fac]})
        return experimental_factors

    def get_condition_name(self):
        experimental_setting = utils.find_position(self.gen.result_dict, self.position[:-3] + ['setting_id'])
        if experimental_setting not in self.gen.conditions:
            factors = utils.find_position(self.gen.result_dict, self.position[:-3] + ['experimental_factors'])
            factor_combis = {}
            for i in range(len(factors)):

                factor_info = \
                    list(utils.find_keys(self.gen.key_yaml, factors[i]['factor']))[0]
                factor = factors[i]['factor']
                values = copy.deepcopy(factors[i]['values'][0]) if len(
                    factors[i]['values']) == 1 and isinstance(
                    factors[i]['values'][0], dict) else copy.deepcopy(
                    factors[i]['values'])
                # if the values of the experimental factor are in a dictionary or the
                # factor contains a list (so the factor can occur multiple times in a
                # condition) than call the function get_combinations to create all
                # possible combinations of this factor with its values
                if factor_info['list'] or (
                        isinstance(factor_info['value'], dict)
                        and not ('special_case' in factor_info
                                 and 'value_unit' in
                                 factor_info['special_case'])):
                    for val in values:
                        val_info = list(utils.find_keys(self.gen.key_yaml, val))
                        if len(val_info) > 0:
                            val_info = val_info[0]
                            if 'special_case' in val_info:
                                if 'merge' in val_info['special_case']:
                                    for j in range(len(values[val])):
                                        new_val = []
                                        for elem_key in values[val][j]:
                                            new_val.append(
                                                f'{elem_key}:"{values[val][j][elem_key]}"')
                                        values[val][
                                            j] = f'{"{"}{"|".join(new_val)}{"}"}'
                        for j in range(len(values[val])):
                            if isinstance(values[val][j], dict):
                                new_val = "|".join([f'{key}:"{values[val][j][key]}"' for key in values[val][j]])
                                values[val][j] = f'{"{"}{new_val}{"}"}'

                    if 'special_case' in factor_info:
                        if 'group' in factor_info['special_case']:
                            values['ident_key'] = factor_info['special_case'][
                                'group']
                        if 'control' in factor_info['special_case']:
                            values['control'] = factor_info['special_case'][
                                'control']
                    # overwrite the values with the combinations
                    factor_combis[factor] = utils.get_combis(
                        values, factor, self.gen.result_dict, self.gen.key_yaml)
                elif 'special_case' in factor_info and 'value_unit' in factor_info['special_case']:
                    values = [f'{val["value"]}{val["unit"]}' for val in values]
                    factor_combis[factor] = [f'{factor}:"{val}"' for val in values]
                else:
                    factor_combis[factor] = [f'{factor}:"{val}"' for val in values]
            factor_dict = []
            for elem_key in factor_combis:
                factor_dict.append(
                    {'factor': elem_key, 'values': factor_combis[elem_key]})
            # call get_condition_combinations to create all conditions
            combinations = utils.get_condition_combinations(factor_dict)
            for key in factor_combis:
                factor_combis[key] = [x for x in factor_combis[key] if x.count(f'{key}:') <= 1]
            self.gen.conditions[experimental_setting] = [factor_combis, combinations]

        i = 1
        print('Please combine the factors to create a condition.')
        for key in self.gen.conditions[experimental_setting][0]:
            print(f'\n\033[1m{key}\033[0m')
            for elem in self.gen.conditions[experimental_setting][0][key]:
                print(f'{i} {elem}')
                i += 1

        w = [x for xs in list(self.gen.conditions[experimental_setting][0].values())
             for x in xs]
        used_values = self.gen.parse_input_list(w, False)
        condition_name = None
        for elem in self.gen.conditions[experimental_setting][1]:
            if all([x in elem for x in used_values]) and len(elem) == len('-'.join(used_values)):
                condition_name = elem
                break
        if condition_name is not None:
            filled_conditions = []
            for i in range(self.position[-2]):
                filled_conditions.append(utils.find_position(self.gen.result_dict, self.position[:-2] + [i, 'condition_name']))
            if condition_name not in filled_conditions:
                return condition_name
            else:
                print('You already selected this condition. Try again.')
                return self.get_condition_name()
        else:
            print('Invalid condition. Please try again.')
            return self.get_condition_name()

    def get_samples(self):
        cond_name = utils.find_position(self.gen.result_dict, self.position[:-2] + ['condition_name'])
        test = utils.split_cond(cond_name)
        sample_structure = list(utils.find_keys(self.gen.key_yaml, 'samples'))[0]
        for elem in test:
            sample_structure['value'][elem[0]]['value'] = elem[1]
            if 'special_case' in sample_structure['value'][elem[0]]:
                sample_structure['value'][elem[0]]['special_case']['factor'] = True
            else:
                sample_structure['value'][elem[0]]['special_case'] = {'factor': True}
        self.gen.parse_lists(sample_structure, self.position, 2, self.gen.result_dict, is_factor=False)
        return utils.find_position(self.gen.result_dict, self.position)

    def get_sample_name(self, read_in_whitelists=None):
        if self.position[-2] == 'technical_replicates':
            '''count = utils.find_position(self.gen.result_dict, self.position[:-1] + ['count'])
            nom = utils.find_position(self.gen.result_dict, self.position[:-2] + ['number_of_measurements'])
            sample_name = utils.find_position(self.gen.result_dict, self.position[:-2] + ['sample_name'])
            organism = utils.find_position(self.gen.result_dict, self.position[:self.position.index('experimental_setting')+2] + ['organism', 'organism_name'])
            samples = []
            for i in range(count):
                # iterate over the number of measurements
                for j in range(nom):
                    # add sample name containing id, organism, sample identifier,
                    # index of technical replicate and index of measurement to samples
                    # list
                    samples.append(f'{self.gen.project_id}_{organism}_{sample_name}_'
                                   f't{"{:02d}".format(i + 1)}_'
                                   f'm{"{:02d}".format(j + 1)}')'''
            return utils.create_sample_names(self.gen.result_dict, {}, self.position)
        else:
            condition = utils.find_position(self.gen.result_dict, self.position[:-4]+['condition_name'])
            return f'{utils.get_short_name(condition, self.gen.result_dict, self.gen.key_yaml, read_in_whitelists=read_in_whitelists)}_b{"{:02d}".format(self.position[-2] + 1)}'

    def get_filenames(self):
        setting_index = self.position.index('experimental_setting') + 1
        experimental_factors = utils.find_position(self.gen.result_dict, self.position[:setting_index+1] + ['experimental_factors'])
        double = []
        if len(list(utils.find_keys(experimental_factors, 'gene'))) > 0 or len(list(utils.find_values(experimental_factors, 'gene'))) > 0:
            gene_whitelist = utils.get_whitelist('gene', self.gen.result_dict)
            double = gene_whitelist['double'] if 'double' in gene_whitelist else []
        return utils.create_filenames(self.gen.result_dict, double, self.position, {})

    def get_count(self):
        return len(utils.find_position(self.gen.result_dict, self.position[:-1] + ['samples']))

    def get_publications(self):
        pubmed_ids = list(utils.find_keys(self.gen.result_dict, 'pubmed_id'))
        if any([x not in self.gen.publications for x in pubmed_ids]):
            records = utils.get_publication_object(pubmed_ids, self.gen.email)
            for record in records:
                if 'Id' in record and int(record['Id']) not in self.gen.publications:
                    self.gen.publications[int(record['Id'])] = record

    def get_author(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = [str(x) for x in self.gen.publications[pubmed_id]['AuthorList'][0:3]] + ['et. al'] \
            if len(self.gen.publications[pubmed_id]['AuthorList']) > 3 \
            else [str(x) for x in self.gen.publications[pubmed_id]['AuthorList']]
        except ValueError:
            res = None
        return res

    def get_title(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = str(self.gen.publications[pubmed_id]['Title'])
        except ValueError:
            res = None
        return res

    def get_journal(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = str(self.gen.publications[pubmed_id]['Source'])
        except ValueError:
            res = None
        return res

    def get_volume(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = int(self.gen.publications[pubmed_id]['Volume'])
        except ValueError:
            res = None
        return res

    def get_year(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = int(self.gen.publications[pubmed_id]['PubDate'].split(' ')[0])
        except ValueError:
            res = None
        return res

    def get_issue(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = int(self.gen.publications[pubmed_id]['Issue'])
        except ValueError:
            res = None
        return res

    def get_pages(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = str(self.gen.publications[pubmed_id]['Pages'])
        except ValueError:
            res = None
        return res

    def get_doi(self):
        self.get_publications()
        pubmed_id = utils.find_position(self.gen.result_dict,
                                        self.position[:-1] + ['pubmed_id'])
        try:
            res = str(self.gen.publications[pubmed_id]['DOI'])
        except ValueError:
            res = None
        return res

    def get_techniques(self):
        all_settings = self.gen.setting_ids
        structure = list(utils.find_keys(self.gen.key_yaml, 'techniques'))[0]
        structure['list'] = False
        for i in range(len(all_settings)):
            self.gen.parse_lists(structure, self.position + [i], 2, self.gen.result_dict)
        return utils.find_position(self.gen.result_dict, self.position)

    def get_setting(self):
        return self.gen.setting_ids[self.position[-2]]






import os

import src.file_reading as file_reading
import src.utils as utils
from src.generate import Generate
from tabulate import tabulate

# This scripts implements functions to find metadata files that contain values
# specified in a search string


# set parameter size to terminal width in order to display printed tables
# correctly
# if that is not possible, set the size to the default value 80
try:
    size = os.get_terminal_size()
except OSError:
    size = 80


def find_projects(key_yaml, dir_path, search, return_dict, skip_validation=False):
    """
    This function iterates through the search string and evaluates all parts
    within round brackets from the inside to the outside. It calls the function
    'get_search_parameters' for every sub string within brackets. It returns
    all matching metadata files in a dictionary.
    :param dir_path: a folder path that is searched for metadata files
    :param search: a search string containing values that should match the
                   metadata
    :param return_dict: bool: if True the whole metafile is returned,
                              if False just the metafile path is returned
    :return: a dictionary containing all matches
             key=input_id, value=dictionary or path depending on return_dict
    """
    # split parameters linked via or into list
    # read in all *_metadata.yaml(yml) within the path
    metafiles, validation_reports = file_reading.iterate_dir_metafiles(
        key_yaml, [dir_path], return_false=True, skip_validation=skip_validation
    )

    # put round brackets around the search string
    search = "(" + search + ")"

    # initialize parameter result with an empty list
    # it is used to store the matching metadata files
    result = []

    # iterate over the read in metadata files
    for metafile in metafiles:

        if "project" in metafile and "id" in metafile["project"]:

            # initialize parameter sub_list and sub with empty string
            # sub stores parts of the search string that is read in
            # whenever a bracket is reached, the string in sub is evaluated and
            # the result is added to sub_list
            sub_list = ""
            sub = ""

            # set brace_count to 0 meaning that there is currently no open
            # brace
            brace_count = 0

            # iterate over search string
            for letter in search:

                # if an opening bracket is reached, add 1 to brace_count, add
                # the string within sub to the sub_list and set sub to an empty
                # string
                if letter == "(":
                    brace_count += 1
                    sub_list += sub
                    sub = ""

                # if a closing bracket is reached, subtract 1 from brace_count
                elif letter == ")":
                    brace_count -= 1

                    # test if the closing bracket is the last bracket in the
                    # search string
                    if brace_count != 0:

                        # if it is the last bracket, add the string in sub to
                        # the sub_list
                        sub_list += sub

                    else:

                        # if it is a bracket within the search string and sub
                        # is not an empty string then use the function
                        # parse_search_parameters to evaluate the string in sub
                        # and add the result to the sub_string
                        if sub != "":
                            res = parse_search_parameters(metafile, sub)
                            sub_list += str(True) if res else str(False)

                    # set sub to an empty string again
                    sub = ""

                else:

                    # add the current character of the search string to sub
                    sub += letter

            # evaluate the sub_list with parse_search_parameters if it does not
            # equal True or False
            if sub_list != "True" and sub_list != "False":
                sub_list = str(parse_search_parameters(metafile, sub_list))

            # add a dictionary containing information about the metadata file
            # to result if it matches the search
            # key: project id
            # value: whole metafile or path to metafile depending on whether
            # result_dict is set to True or False
            if sub_list == "True":
                result.append(
                    {
                        metafile["project"]["id"]: (
                            metafile["path"] if return_dict == False else metafile
                        )
                    }
                )

    # return the list of matching metadata files
    return result


def print_summary(result, output):
    """
    This function prints information about the metadata files that match the
    search string in a table. The information includes the project id, path to
    the metadata file, project name and name of the owner.
    :param result: a list containing dictionaries with project id as key and
                   metadata dictionary as value for every metadata file that
                   matches the search string
    :return: res: a table containing id, path, name and owner of every project
                  whose metadata matches the search string
    """

    #  initialize res as a list and define the headers of the table as the
    #  first list element
    res = [["ID", "Path", "Project name", "Owner"]]
    save_res = []
    # iterate over dictionaries for matching metadata files
    for elem in result:
        # iterate over metadata IDs
        for key in elem:
            project_res = {'id': key}
            try:
                project_path = elem[key]["path"]
                project_res['path'] = project_path
            except KeyError:
                project_path = "Not found"
                project_res['path'] = None
            try:
                project_name = elem[key]["project"]["project_name"]
                project_res['project_name'] = project_name
            except KeyError:
                project_name = "Not found"
                project_res['project_name'] = None
            try:
                owner_name = elem[key]["project"]["owner"]["name"]
                project_res['owner'] = owner_name
            except KeyError:
                owner_name = "Not found"
                project_res['owner'] = None

            try:
                project_res['email'] = elem[key]['project']['owner']['email']
            except KeyError:
                project_res['email'] = None

            project_res["organisms"] = list(set(list(utils.find_keys(elem[key], "organism_name"))))

            try:
                project_res["description"] = elem[key]["project"]["description"]
            except KeyError:
                project_res["description"] = None

            try:
                project_res["date"] = elem[key]["project"]["date"]
            except KeyError:
                project_res["date"] = None

            if "nerd" in elem[key]["project"]:
                nerds = []
                for nerd in elem[key]["project"]["nerd"]:
                    nerds.append(nerd["name"])
                project_res["nerd"] = list(set(nerds))
            else:
                project_res["nerd"] = None

            try:
                techniques = elem[key]['technical_details']['techniques']
                tech_list = []
                for tech in techniques:
                    tech_list += tech['technique']
                project_res['technique'] = list(set(tech_list))
            except:
                project_res['technique'] = None
            
            cell_type = list(set(utils.find_keys(elem[key], 'cell_type')))
            project_res['cell_type'] = cell_type

            tissue = []
            
            tissues = list(utils.find_keys(elem[key], 'tissue'))
            for tis in tissues:
                tissue += tis
            tissue = list(set(tissue))
            project_res['tissue'] = tissue

            # treatment
            treatment = []
            
            medical = list(utils.find_list_key(elem[key], 'medical_treatment:treatment_type'))
            treatment += list(set(medical))
            
            physical = list(utils.find_keys(elem[key], 'physical_treatment'))
            treatment += list(set(physical))
            
            injury = list(utils.find_list_key(elem[key], 'injury:injury_type'))
            treatment += list(set(injury))

            project_res['treatment'] = treatment

            # disease

            disease = list(utils.find_list_key(elem[key], 'disease:disease_type'))
            project_res['disease'] = list(set(disease))

                # add the id, path, project_name and owner to res
            res.append([key, project_path, project_name, owner_name])
            save_res.append(project_res)

    if output == 'print':
        # convert res into a table using tabulate
        # set format to fancy_grid, define the headers as the first row and set the
        # column width to fit to the terminal size
        res = tabulate(
            res,
            tablefmt="fancy_grid",
            headers="firstrow",
            maxcolwidths=[
                size.columns * 1 / 8,
                size.columns * 3 / 8,
                size.columns * 3 / 8,
                size.columns * 1 / 8,
            ],
        )
    elif output == 'json':
        res = save_res
    # return the table
    return res


def parse_search_parameters(metafile, search):
    """
    This function splits the search parameters at 'or' and stores them in a
    list. Every element within this list is then additionally split at 'and'
    resulting in a nested list. The function 'get_matches' is the called with
    the metafile and the nested search parameters.
    :param metafile: a dictionary containing information of a metadata file
    :param search: the search string containing the values to search for
    :return: result: True if there is a match, else False
    """

    # split the search string at 'or' and store the parts of the search string
    # in a list
    search_parameters = search.split(" or ")

    # iterate over the parts of the search string that were split at or
    for i in range(len(search_parameters)):

        # split parameters linked via 'and' within the 'or-list' -> nested list
        search_parameters[i] = search_parameters[i].split(" and ")

        # iterate over the parts of the search string split at 'and'
        for j in range(len(search_parameters[i])):

            # test if the part is not equal to True or False
            if search_parameters[i][j] != "True" or "False":

                # call get_should_be_in to append True or False to the search
                # parameter separated by ':' (depending on whether there is a
                # 'not' in front of the search parameter)
                search_parameters[i][j] = get_should_be_in(search_parameters[i][j])

    # call function get_matches to find matches in the nested list for 'or'
    # and 'and'
    results = get_matches(metafile, search_parameters)

    # if there is at least one True then there is a match (because the value
    # was found in at least one metadata field) and the result is set to True,
    # otherwise the result is set to False
    if True in results:
        result = True
    else:
        result = False

    # return True if there is a match else False
    return result


def get_matches(metafile, search_parameters):
    """
    This function takes a metafile and is recursively called on every element
    of the metafile if the metafile is a list. Otherwise, it calls the function
    'calculate_match' on the metafile to test if the search parameters occur in
    the metadata.
    :param metafile: a dictionary containing information of a metadata file
    :param search_parameters: a nested list containing all search parameters
                              chained by 'or' (outer list) and 'and'
                              (inner lists)
    :return: result: a list containing all matches as boolean values (True and
                     False)
    """

    # initialize an empty list to store the results
    results = []

    # test if the metafile is a list
    if isinstance(metafile, list):

        # iterate over elements in  metadata list
        for x in metafile:

            # call this function for every element
            results += get_matches(x, search_parameters)

    else:

        # if the metafile is not a list call calculate_match for it and add it
        # to result
        results += [calculate_match(metafile, search_parameters)]

    # return the list containing all matches (True or False)
    return results


def calculate_match(metafile, search_parameters):
    """
    This function iterated through all search values that were chained by 'and'
    or 'or' and calls the function 'find_entry' on them to test whether the
    value occurs in the metadata file.
    :param metafile: a dictionary containing information of a metadata yaml
    :param search_parameters: a nested dictionary containing all parameters
                              split at 'or' (outer list) and 'and'
                              (inner lists)
    :return: True if a match was found else False
    """

    # initialize a list to store all matches within the outer list containing
    # parameters that were chained by 'or'
    or_found = []

    # iterate through outer list -> or
    for or_param in search_parameters:

        # initialize a list to store all matches within the inner list
        # containing parameters that were chained by 'and'
        and_found = []

        # iterate through inner list -> and
        for and_param in or_param:

            # split search parameter at ':'
            # last element in list saved in 'should-be_found'
            # -> False if 'not' was specified for the parameter

            # test if the search parameter contains double quotes
            if '"' in and_param:

                # initialize a list to store
                params = []

                # split the search parameter at ':'
                # the last element of the emerging list declares if the
                # parameter should be found in the metafile and is saved in
                # should_be_found and removed from the search parameter
                should_be_found = and_param.split(":")[-1]
                and_param = and_param.rstrip(f":{should_be_found}")

                # split the search parameter into keys and value
                p = and_param.rstrip('"').split('"')

                # if there is a value for the keys than split it at ':' to get
                # a list of all chained metadata keys and save them in 'params'
                if p[0] != "":
                    params = p[0].rstrip(":").split(":")

                # add the value to the 'params' list
                params.append(p[1])

                # add 'should_be_found' to the 'params' list
                params.append(should_be_found)
            else:

                # if there is no double quote in the search parameter then
                # split ist at ':' and save it into 'params'
                # assign the value of the last element to 'should_be_found'
                params = and_param.split(":")
                should_be_found = params[-1]

            # if there are no keys specified and the value was already
            # evaluated (e.g. as part of a sub search string within brackets)
            # then set the match to True if the value equals 'True' and to
            # False if it equals 'False'
            if len(params) == 1 and params[0] in ["True", "False"]:
                match = bool(params[0])
            elif len(params) == 2 and params[0] == "True":
                match = True
            elif len(params) == 2 and params[0] == "False":
                match = False

            else:

                # call find_entry to evaluate match
                match = find_entry(metafile, params[0:-2], params[-2])

            # set match to True if it was found while it was supposed to be
            # found or if it was not found while not being supposed to be found
            # otherwise set match to False
            if (match and should_be_found == "True") or (
                not match and should_be_found == "False"
            ):
                and_found.append(True)
            else:
                and_found.append(False)

        # append 'True' to the or_found list if all parameters in the inner
        # list for 'and' are True, otherwise append 'False'
        if False not in and_found:
            or_found.append(True)
        else:
            or_found.append(False)

    # return True if at least one parameter chained by 'or' is True, otherwise
    # return False
    return True if True in or_found else False


def get_should_be_in(param):
    """
    This function tests if a 'not' was stated in front of a search value and
    saves this information into the variable 'should_be_found'. The 'not' gets
    removed from the search value and 'should_be_found' is appended at its end
    divided by ':'.
    :param param: a search parameter containing a value that can be chained
                  with keys (e.g. key1:key2:"value") and can contain a 'not'
    :return: p: the search parameter with the appended value for
                'should_be_found' and the removed 'not'
    """

    # test if the given parameter is a list
    if isinstance(param, list):

        # initialize a list to store results for every list element
        p = []

        # iterate over list elements
        for x in param:

            # call this function on the element and append the result to p
            p.append(get_should_be_in(x))

    # if the parameter is not of type list
    else:

        # set 'should_be_in' to True as a default value
        should_be_in = True

        # test if 'not' is present in the parameter
        if "not " in param:

            # set 'should_be_in' to False (if 'not' was found)
            should_be_in = False

            # replace 'not' with empty string
            param = param.replace("not ", "")

        # strip whitespaces from the parameter and append 'should_be_in' at the
        # end divided by ':'
        p = param.strip() + (":" + str(should_be_in))

    # return the parameter containing 'should_be_in'
    return p


def find_entry(metafile, targets, target_value):
    """
    This function tests if a value (and key) is found in a metadata dictionary.
    :param metafile: dictionary containing information of metadata yaml
    :param targets: list of keys in order of depth in yaml
                    -> [key1, key2, key3] for 'key1:key2:key3:value'
    :param target_value: the value that should be found
    :return: True if match was found, else False
    """

    # test if keys were specified to restrict the metadata fields that should
    # be searched
    if len(targets) >= 1:

        # initialize the result with the metafile
        result = [metafile]

        # iterate over the keys
        for key in targets:

            # initialize 'r2' with an empty list
            # this parameter is used to store part results for every element in
            # 'result'
            r2 = []

            # iterate over elements in 'result'
            for item in result:

                # call find_keys from the script utils.py for the current key
                # and result item to get a list of all occurrences of this key
                # within the item
                r2 += list(utils.find_keys(item, key))

            # set the result to the part result
            # this way the metadata fields get restricted because just the
            # parts with matching keys are used
            result = r2

    # iterate through whole dictionary if no key was specified in 'targets'
    else:
        result = list(utils.find_values(metafile, target_value))

    # iterate over the results
    for value in result:

        # test if the values match if they are of type int or boolean and
        # return True if that is the case
        if (type(target_value) is int and type(value) is int) or (
            type(target_value) is bool and type(value) is bool
        ):
            if target_value == value:
                return True

        # test if the search value is a substring of the metadata value if they
        # are of type string and return True if that is the case
        else:
            if all(
                elem in str(value).lower()
                for elem in str(target_value).lower().split(" ")
            ):
                return True

    # return False if no match was found
    return False



import copy

import src.web_interface.whitelist_parsing as whitelist_parsing
import src.web_interface.wi_utils as wi_utils
import datetime
import src.utils as utils

# TODO: example of format in wi_object (Keys)


def get_empty_wi_object(key_yaml, read_in_whitelists):
    """
    This function parses the keys.yaml and returns an empty object for the web
    interface.
    :return: wi_object: object containing all information from the keys.yaml
             in a format readable by the web interface
    """

    # initialize emtpy dict to save wi_object to
    wi_object = {}

    # iterate over all keys in general structure
    for key in key_yaml:

        # parse information of a key into wi_object format
        wi_object[key], whitelist_object = parse_empty(key_yaml[key], key,
                                                       key_yaml, {}, read_in_whitelists)

    # add a key 'all_factors' with an empty list as value to the object
    # the experimental factors will be saved here after input on the website
    wi_object['all_factors'] = []
    wi_object['publication_records'] = {}

    return wi_object


def parse_empty(node, pos, key_yaml, filled_object, read_in_whitelists,
                get_whitelist_object=False):
    """
    This function parses a part of the key.yaml into an object readable by the
    web-interface
    :param filled_object: a dictionary containing pre-filled information
    :param node: a part of the key.yaml that should be parsed
    :param pos: the position of the node (chained keys)
    :param key_yaml: the whole key.yaml
    :param get_whitelist_object: bool, True if whitelists should be written
                                 into a separate object, False if not, default
                                 False
    :return: part_object: an object for the web interface parsed from node
             whitelist_object: a dictionary containing all whitelists, it is
                               only filled if 'get_whitelist_object' is set to
                               True
    """

    # disable input for the keys condition_name and sample_name because they
    # are generated
    input_disabled = True if 'special_case' in node and 'generated' in \
                             node['special_case'] and \
                             node['special_case']['generated'] in \
                             ['end', 'now'] and pos.split(':')[-1] not in ['id', 'setting', 'techniques'] \
        else False

    # initialize a whitelist object to store all whitelists
    whitelist_object = {}

    # test if the value of the current node is a dictionary
    if isinstance(node['value'], dict):
        # initialize the input type with None
        input_type = None

        # special case: merge or value_unit
        if 'special_case' in node and ('merge' in node['special_case'] or
                                       'value_unit' in node['special_case']):

            # read and parse whitelist
            whitelist, whitelist_type, input_type, headers, whitelist_keys, double = \
                whitelist_parsing.parse_whitelist(pos.split(':')[-1], node,
                                                  filled_object, whitelist_object=read_in_whitelists)

            # set input type of field organism to 'organism' -> special case
            # (used to display button to confirm organism on the website)
            if pos.split(':')[-1] == 'organism':
                input_type = 'organism_name'

            # test if the whitelist should be stored in a separate object
            if get_whitelist_object:

                # special case: value_unit
                if input_type == 'value_unit':

                    # store the whitelist in the object under the key 'unit'
                    whitelist_object['unit'] = whitelist
                    whitelist = 'unit'

                # a whitelist is defined for the key
                elif whitelist is not None:

                    # save the whitelist to the object under the keyname
                    whitelist_object[pos.split(':')[-1]] = whitelist
                    whitelist = pos.split(':')[-1]

            # creation and filling of dictionary containing all necessary
            # information for one input field
            part_object = {'position': pos, 'mandatory': node['mandatory'],
                           'list': node['list'],
                           'displayName': node['display_name'],
                           'desc': utils.print_desc(node['desc'], 'html').replace('\n', '<br>'),
                           'value': None,
                           'whitelist': whitelist,
                           'whitelist_type': whitelist_type,
                           'input_type': input_type,
                           'input_disabled': input_disabled,
                           'delete_disabled': False}

            if (part_object['whitelist'] == ['True', 'False'] or part_object['whitelist'] == 'pooled') and part_object[
                'value'] is not \
                    None:
                if part_object['value']:
                    part_object['value'] = 'True'
                else:
                    part_object['value'] = 'False'
            # special case : value unit -> add key value_unit to dict
            if input_type == 'value_unit':
                part_object['value_unit'] = node['value']['unit']['value']

            # save headers and whitelist keys in the object if they are defined
            if headers is not None:
                part_object['headers'] = headers
            if whitelist_keys is not None:
                part_object['whitelist_keys'] = whitelist_keys

            # add a search info for autofill
            if input_type in ['single_autofill', 'multi_autofill']:
                part_object['search_info'] = {
                    'organism': filled_object['organism']
                    if 'organism' in filled_object else None,
                    'key_name': part_object['position'].split(':')[-1]}

        # no special case -> the value takes a dictionary and should be
        # displayed via an expandable
        else:

            # initialize variable input_fields with empty list
            # it is used to save the keys of the 'value' dictionary as input
            # fields
            input_fields = []

            # iterate over all keys in value
            for key in node['value']:

                if not ('special_case' in node['value'][key] and 'invisible' in
                        node['value'][key]['special_case'] and
                        node['value'][key]['special_case']['invisible']):

                    # call this function to create a dictionary object for all keys
                    # storing their necessary information
                    field_infos, w_object = parse_empty(
                        node['value'][key], pos + ':' + key, key_yaml,
                        filled_object, read_in_whitelists,
                        get_whitelist_object=get_whitelist_object)

                    # add the parsed input fields to the list
                    input_fields.append(field_infos)

                    # merge the whitelist object with the object from the input
                    # field
                    whitelist_object = {**whitelist_object, **w_object}

            # creation and filling of dictionary containing all necessary
            # information for one expandable with its input fields
            part_object = {'position': pos, 'mandatory': node['mandatory'],
                           'list': node['list'], 'title': node['display_name'],
                           'desc': utils.print_desc(node['desc'], 'html').replace('\n', '<br>'), 'input_fields': input_fields,
                           'original_input': copy.deepcopy(input_fields),
                           'input_disabled': input_disabled,
                           'delete_disabled': False}

            if 'special_case' in node:
                if 'group' in node['special_case']:
                    part_object['ident_key'] = f'{pos}:{node["special_case"]["group"]}'
                if 'control' in node['special_case']:
                    part_object['control'] = node['special_case']['control']

            if 'short_desc' in node:
                part_object['short_desc'] = node['short_desc']
        # test if the key takes multiple values and add the property
        # 'list_value' as a place to save those values to via the website
        if node['list'] or input_type == 'single_autofill':
            part_object['list_value'] = []

    # the key does not contain a dictionary as value
    else:
        # read and parse whitelist
        whitelist, whitelist_type, input_type, headers, whitelist_keys, double = \
            whitelist_parsing.parse_whitelist(pos.split(':')[-1], node,
                                              filled_object, whitelist_object=read_in_whitelists)

        # test if whitelists are defined and should be stored in a separate
        # object
        if get_whitelist_object and whitelist is not None:

            # set the whitelist to the name of the key if a whitelist was
            # defined
            whitelist_object[pos.split(':')[-1]] = whitelist
            whitelist = pos.split(':')[-1]

        # creation and filling of dictionary containing all necessary
        # information for one input field
        part_object = {'position': pos, 'mandatory': node['mandatory'],
                       'list': node['list'],
                       'displayName': node['display_name'],
                       'desc': utils.print_desc(node['desc'], 'html').replace('\n', '<br>'), 'value': node['value'],
                       'whitelist': whitelist,
                       'whitelist_type': whitelist_type,
                       'input_type': input_type,
                       'input_disabled': input_disabled,
                       'delete_disabled': False}

        if (part_object['whitelist'] == ['True', 'False'] or part_object['whitelist'] == 'pooled') and part_object['value'] is not \
                None:
            if part_object['value']:
                part_object['value'] = 'True'
            else:
                part_object['value'] = 'False'

        # add headers and whitelist keys to the empty object if they are
        # defined
        if headers is not None:
            part_object['headers'] = headers
        if whitelist_keys is not None:
            part_object['whitelist_keys'] = whitelist_keys

        # test if the key takes multiple values or uses autofill and add the
        # property 'list_value' as a place to save those values to via the
        # website
        if node['list'] or input_type in ['single_autofill', 'restricted_short_text']:
            part_object['list_value'] = []

        if input_type == 'restricted_short_text':

            # default
            part_object['restriction'] = {'regex': ' ',
                                          'max_length': 10}

            if 'special_case' in node and 'restriction' in node['special_case']:
                if 'regex' in node['special_case']['restriction']:
                    part_object['restriction']['regex'] = \
                        node['special_case']['restriction']['regex']
                if 'max_length' in node['special_case']['restriction']:
                    part_object['restriction']['max_length'] = \
                        node['special_case']['restriction']['max_length']

        # add the key 'search_info' to all fields using autofill to help get
        # the correct whitelist via the website
        if input_type == 'single_autofill' or input_type == 'multi_autofill':
            part_object['search_info'] = {
                'organism': filled_object['organism']
                if 'organism' in filled_object else None,
                'key_name': part_object['position'].split(':')[-1]}

        if input_type == 'date':
            part_object['value'] = wi_utils.str_to_date(datetime.datetime.now().strftime('%d.%m.%Y'))

    return part_object, whitelist_object



import os
import git
import re

#TODO: username, pw
def get_whitelists(whitelist_path, whitelist_repo, whitelist_branch, update_whitelist):
    """
    This function clones the whitelist repository. If the repository already
    exists then it pulls the most recent changes
    """
    if not os.path.exists(whitelist_path) or update_whitelist:

        print('Fetching whitelists...\n')

        if not os.path.exists(whitelist_path):
            repo = git.Repo.clone_from(whitelist_repo,
                                   whitelist_path)
            whitelist_branch = get_branch(repo, whitelist_branch)
            repo.git.checkout(whitelist_branch)
        else:
            repo = git.Repo(whitelist_path)
            repo.remotes.origin.fetch(prune=True, prune_tags=True)
            whitelist_branch = get_branch(repo, whitelist_branch)
            repo.git.checkout(whitelist_branch)
            repo.remotes.origin.pull(whitelist_branch)

        print(f'Fetched branch {whitelist_branch}.\n')
    else:
        repo = git.Repo(whitelist_path)
        repo.remotes.origin.fetch(prune=True, prune_tags=True)
        whitelist_branch = get_branch(repo, whitelist_branch)
    return whitelist_branch


def get_branch(repo, whitelist_branch):
    if '*' in whitelist_branch:
        regex_branch = whitelist_branch.replace(".", "\.").replace("$", "\$").replace("+", "\+").replace("*", ".*")
        branch_list = [a.name.replace('origin/', '') for a in repo.tags + repo.remote().refs if a.name != 'origin/HEAD']
        matching_branches = [x.lstrip('v').split('.') for x in branch_list if re.search(f'^{regex_branch}$', x) is not None]
        splitted_branches = []
        for branch in matching_branches:
            try:
                splitted_branches.append([int(x) for x in branch])
            except ValueError:
                pass
        splitted_branches.sort(reverse=True)
        whitelist_branch = f'v{".".join([str(s) for s in splitted_branches[0]])}'
    return whitelist_branch


import copy
import multiprocessing
import os
from functools import partial
from multiprocessing.pool import Pool
import time
from src import validate_yaml
from src.utils import read_in_yaml
import yaml as yml


# The following functions were inspired by Mampok and slightly customized
# https://gitlab.gwdg.de/loosolab/software/mampok/-/blob/master/mampok/
# file_reading.py


def iterate_dir_metafiles(
    key_yaml,
    path_metafiles,
    filename="_metadata",
    logical_validation=True,
    yaml=None,
    whitelist_path=None,
    return_false=False,
    skip_validation = False,
    show_logs = False
):
    """
    iterate through a list of paths to find all _metadata.yaml(yml) files
    :param path_metafiles: list of paths containing yaml files
    :return: metafile_list: list of dictionaries containing information from
             found metadata files
    """
    start = time.time()
    # turn string into list
    if isinstance(path_metafiles, str):
        path_metafiles = [path_metafiles]

    # iterate through paths, through directories, through files
    metafile_list = []
    file_reports = []
    error_count = 0
    warning_count = 0
    corrupt_count = 0
    items = []
    for path_metafile in path_metafiles:
        items += [
            os.path.join(subdir, file)
            for subdir, dirs, files in os.walk(path_metafile)
            for file in files
            if file.lower().endswith(f"{filename}.yaml")
            or file.lower().endswith(f"{filename}.yml")
        ]
    end_listing = time.time()
    if show_logs:
        print(f'File listing took {"%.2f" % (end_listing-start)} seconds.')
    cpu_count = multiprocessing.cpu_count()
    pool = Pool(cpu_count)
    results = pool.map(
        partial(
            validate,
            filename=filename,
            key_yaml=key_yaml,
            logical_validation=logical_validation,
            whitelist_path=whitelist_path,
            yaml=copy.deepcopy(key_yaml),
            skip_validation=skip_validation,
        ),
        items,
    )
    pool.close()
    end_reading = time.time()
    if show_logs:
        print(f'File reading and validation took {"%.2f" % (end_reading-end_listing)} seconds with {cpu_count} CPUs.')
    reading_times = []
    for result in results:
        if result[0] is not None:
            if result[3] == 0 or return_false:
                metafile_list.append(result[0])

            if result[1]:
                corrupt_count += 1
                file_reports.append(
                    {"file": result[0], "error": result[2], "warning": result[4]}
                )

            error_count += result[3]
            warning_count += result[5]
        reading_times.append(result[6])
    end_result = time.time()
    if show_logs:
        print(f'Parsing the results took {"%.2f" % (end_result-end_reading)} seconds.')
        print(f'File reading per file took MIN {"%.2f" % (min(reading_times))}s | MEAN {"%.2f" % (sum(reading_times)/len(reading_times))}s | MAX {"%.2f" % (max(reading_times))}s')
        try:
            largest_file = metafile_list[reading_times.index(max(reading_times))]
            print(f'Reading took longest for file {largest_file["project"]["id"]}')
        except KeyError:
            pass
        except IndexError:
            pass
    return metafile_list, {
        "all_files": len(results),
        "corrupt_files": {"count": corrupt_count, "report": file_reports},
        "error_count": error_count,
        "warning_count": warning_count,
    }


def validate(ypath, filename, key_yaml, logical_validation, whitelist_path, yaml, skip_validation=False):
    error_reports = None
    warning_reports = None
    warning_count = False
    error_count = 0
    corrupted = False
    report = {}
    # add files with suffix '_metadata.y(a)ml'
    start = time.time()
    try:
        metafile = read_in_yaml(ypath)
    except yml.scanner.ScannerError:
        metafile = None
    end_read = time.time()
    read_time = end_read - start
    # test if metafile is valid
    if metafile is not None:
        if not skip_validation:
            (
                valid,
                missing_mandatory_keys,
                invalid_keys,
                invalid_entries,
                invalid_values,
                logical_warn,
            ) = validate_yaml.validate_file(
                metafile,
                key_yaml,
                filename,
                logical_validation=logical_validation,
                yaml=yaml,
                whitelist_path=whitelist_path,
            )
            end_val = time.time()
            #print(f'The validation of ONE file took {"%.2f" % (end_val-end_read)} seconds.')
            # add path to dic
            if not valid:
                error_reports = (
                    missing_mandatory_keys,
                    invalid_keys,
                    invalid_entries,
                    invalid_values,
                )
                corrupted = True
                error_count += (
                    len(missing_mandatory_keys)
                    + len(invalid_keys)
                    + len(invalid_entries)
                    + len(invalid_values)
                )

            if len(logical_warn) > 0:
                corrupted = True
                warning_count += len(logical_warn)
                warning_reports = logical_warn
            #print(f"validated file {ypath}")

        metafile["path"] = ypath
    return (
        metafile,
        corrupted,
        error_reports,
        error_count,
        warning_reports,
        warning_count,
        read_time
    )
